"""
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------

solve

   a  + u a  + v a  = 0
    t      x      y

"""

from util import runparams
from numarray import *
import reconstructionlib

def unsplitFluxes(nx, ny, ng, dt, x, y, a):

    imin = ng
    imax = ng+nx-1

    jmin = ng
    jmax = ng+ny-1

    qx = 2*ng+nx
    qy = 2*ng+ny

    # get the advection velocities
    u = runparams.getParam("advection.u")
    v = runparams.getParam("advection.v")

    dx = x[2] - x[1]
    dy = y[2] - y[1]
    
    cx = u*dt/dx
    cy = v*dt/dy
    

    #--------------------------------------------------------------------------
    # monotonized central differences
    #--------------------------------------------------------------------------

    limiter = runparams.getParam("advection.limiter")
    ldelta_a = zeros((qx,qy), Float64)
    ldelta_a = reconstructionlib.limita(1, limiter, nx, ny, ng, a)


    # in the pure advection case, there is no Riemann problem we need to
    # solve -- we just simply do upwinding.  So there is only one 'state'
    # at each interface, and the zone the information comes from depends
    # on the sign of the velocity -- this should be vectorized.
    a_x = zeros((qx,qy), Float64)
    
    """
    
    the fluxes are going to be defined on the left edge of the
    computational zones

            
     |             |             |             |
     |             |             |             |
    -+------+------+------+------+------+------+--
     |     i-1     |      i      |     i+1     |
    
              a_l,i  a_r,i   a_l,i+1
            
    
    a_r,i and a_l,i+1 are computed using the information in
    zone i,j.
    
    """

    if (u < 0):
        # a_x[i,j] = a[i,j] - 0.5*(1.0 + cx)*ldelta_a[i,j]
        a_x[:,:] = a[:,:] - 0.5*(1.0 + cx)*ldelta_a[:,:]
    else:
        # a_x[i,j] = a[i-1,j] + 0.5*(1.0 - cx)*ldelta_a[i-1,j]        
        a_x[1:,:] = a[0:qx-1,:] + 0.5*(1.0 - cx)*ldelta_a[0:qx-1,:]


    # y-direction
    ldelta_a[:,:] = 0.0
    ldelta_a = reconstructionlib.limita(2, limiter, nx, ny, ng, a)

    a_y = zeros((qx,qy), Float64)

    
    # the fluxes are going to be defined on the left edge of the
    # computational zones

    if (v < 0):
        # a_y[i,j] = a[i,j] - 0.5*(1.0 + cy)*ldelta_a[i,j]
        a_y[:,:] = a[:,:] - 0.5*(1.0 + cy)*ldelta_a[:,:]
    else:
        # a_y[i,j] = a[i,j-1] + 0.5*(1.0 - cy)*ldelta_a[i,j-1]
        a_y[:,1:] = a[:,0:qy-1] + 0.5*(1.0 - cy)*ldelta_a[:,0:qy-1]


    # compute the transverse flux differences.  The flux is just (u a)
    # HOTF
    F_xt = u*a_x
    F_yt = v*a_y

    F_x = zeros((qx,qy), Float64)
    F_y = zeros((qx,qy), Float64)

    # the zone where we grab the transverse flux derivative from
    # depends on the sign of the advective velocity
    
    if u <= 0:
        mx = 0
    else:
        mx = -1


    if v <= 0:
        my = 0
    else:
        my = -1


    dtdx2 = 0.5*dt/dx
    dtdy2 = 0.5*dt/dy
        
    i = imin
    while (i <= imax+1):

        j = jmin
        while (j <= jmax+1):

            F_x[i,j] = u*(a_x[i,j] - dtdy2*(F_yt[i+mx,j+1] - F_yt[i+mx,j]))
            
            j += 1
        i += 1

    i = imin
    while (i <= imax+1):

        j = jmin
        while (j <= jmax+1):

            F_y[i,j] = v*(a_y[i,j] - dtdx2*(F_xt[i+1,j+my] - F_xt[i,j+my]))
            
            j += 1
        i += 1
        
            
    return [F_x, F_y]


    

# a split implementation of the Colella method -- just project the multi-d
# equations along a coordinate axis and do the same thing -- this time,
# there is no transverse flux derivative explicitly incorporated.

def splitFluxes(dir, nx, ny, ng, dt, x, y, a):

    imin = ng
    imax = ng+nx-1

    jmin = ng
    jmax = ng+ny-1

    qx = 2*ng+nx
    qy = 2*ng+ny

    # get the advection velocities
    u = runparams.getParam("advection.u")
    v = runparams.getParam("advection.v")

    dx = x[2] - x[1]
    dy = y[2] - y[1]
    
    cx = u*dt/dx
    cy = v*dt/dy
    

    #--------------------------------------------------------------------------
    # monotonized central differences
    #--------------------------------------------------------------------------

    limiter = runparams.getParam("advection.limiter")
    ldelta_a = zeros((qx,qy), Float64)

    if dir == "x":
        
        ldelta_a = reconstructionlib.limita(1,limiter,  nx, ny, ng, a)
                                
        # in the pure advection case, there is no Riemann problem we need
        # to solve -- we just simply do upwinding.  So there is only one
        # 'state' at each interface, and the zone the information comes
        # from depends on the sign of the velocity
        a_x = zeros((qx,qy), Float64)

        """
        
        the fluxes are going to be defined on the left edge of the
        computational zones

        
        |             |             |             |
        |             |             |             |
        -+------+------+------+------+------+------+--
        |     i-1     |      i      |     i+1     |
                
                a_l,i  a_r,i   a_l,i+1
                
                
        a_r,i and a_l,i+1 are computed using the information in
        zone i,j.
            
        """

        if (u < 0):
            # a_x[i,j] = a[i,j] - 0.5*(1.0 + cx)*ldelta_a[i,j]
            a_x[:,:] = a[:,:] - 0.5*(1.0 + cx)*ldelta_a[:,:]
        else:
            # a_x[i,j] = a[i-1,j] + 0.5*(1.0 - cx)*ldelta_a[i-1,j]
            a_x[1:,:] = a[0:qx-1,:] + 0.5*(1.0 - cx)*ldelta_a[0:qx-1,:]

        F_x = zeros((qx,qy), Float64)
            

    elif dir == "y":
                                
        # y-direction
        ldelta_a = reconstructionlib.limita(2, limiter, nx, ny, ng, a)

        a_y = zeros((qx,qy), Float64)

        # the fluxes are going to be defined on the left edge of the
        # computational zones

        if (v < 0):
            # a_y[i,j] = a[i,j] - 0.5*(1.0 + cy)*ldelta_a[i,j]
            a_y[:,:] = a[:,:] - 0.5*(1.0 + cy)*ldelta_a[:,:]
        else:
            # a_y[i,j] = a[i,j-1] + 0.5*(1.0 - cy)*ldelta_a[i,j-1]
            a_y[:,1:] = a[:,0:qy-1] + 0.5*(1.0 - cy)*ldelta_a[:,0:qy-1]
            
        F_y = zeros((qx,qy), Float64)

      
    if dir == "x":
        F_x = u*a_x
        return F_x

    elif dir == "y":
        F_y = v*a_y
        return F_y


            
    
