"""
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
"""

from util import runparams
from mesh import *
from numarray import *
from advectiveFluxes import *

SMALL = 1.e-12


def initialize(myPatch):
    """ initialize the variables for advection """

    # start by making sure that grid is a valid grid object
    if not isinstance(myPatch, patch.cellCentered):
        print "ERROR: patch invalid in io.dumpPatch()"
        print myPatch.__class__
        return -1                          

    myPatch.registerVar("density")

                                                                                    

def evolve(myPatch):
    """ evolve the advection equations through one timestep """


    
    dens = myPatch.getVar("density")

    splitting = runparams.getParam("driver.splitting")

    qx = 2*myPatch.ng + myPatch.nx
    qy = 2*myPatch.ng + myPatch.ny

    dtdx = myPatch.dt/myPatch.dx
    dtdy = myPatch.dt/myPatch.dy
        
    if splitting == "unsplit":

        [flux_x, flux_y] =  unsplitFluxes(myPatch.nx, myPatch.ny, \
                                          myPatch.ng, \
                                          myPatch.dt, myPatch.x, myPatch.y, \
                                          dens)

        """
        do the differencing for the fluxes now.  Here, we use slices so we
        avoid slow loops in python.  This is equivalent to:

        myPatch.data[i,j] = myPatch.data[i,j] + \
                            dtdx*(flux_x[i,j] - flux_x[i+1,j]) + \
                            dtdy*(flux_y[i,j] - flux_y[i,j+1])
        """

        myPatch.data[0,0:qx-1,0:qy-1] = myPatch.data[0,0:qx-1,0:qy-1] + \
                                        dtdx*(flux_x[0:qx-1,0:qy-1] - \
                                              flux_x[1:qx,0:qy-1]) + \
                                        dtdy*(flux_y[0:qx-1,0:qy-1] - \
                                              flux_y[0:qx-1,1:qy])
            
        return 1

    
    elif splitting == "strang" or splitting == "Strang":

        # when we Strang-split, we want to do X(dt) Y(dt) Y(dt) X(dt) a,
        # to advance a two timesteps.  We do X Y in even steps, and
        # Y X in odd steps to get the right symmetry.
        
        if myPatch.nstep % 2 == 0:

            flux_x = splitFluxes('x',
                                 myPatch.nx, myPatch.ny, \
                                 myPatch.ng, \
                                 myPatch.dt, myPatch.x, myPatch.y, \
                                 dens)

            myPatch.data[0,0:qx-1,:] = myPatch.data[0,0:qx-1,:] + \
                                       dtdx*(flux_x[0:qx-1,:] - flux_x[1:qx,:])


            myPatch.fillBC()

            flux_y = splitFluxes('y',
                                 myPatch.nx, myPatch.ny, \
                                 myPatch.ng, \
                                 myPatch.dt, myPatch.x, myPatch.y, \
                                 dens)

            myPatch.data[0,:,0:qy-1] = myPatch.data[0,:,0:qy-1] + \
                                       dtdy*(flux_y[:,0:qy-1] - flux_y[:,1:qy])

        else:
            
            flux_y = splitFluxes('y',
                                 myPatch.nx, myPatch.ny, \
                                 myPatch.ng, \
                                 myPatch.dt, myPatch.x, myPatch.y, \
                                 dens)

            myPatch.data[0,:,0:qy-1] = myPatch.data[0,:,0:qy-1] + \
                                       dtdy*(flux_y[:,0:qy-1] - flux_y[:,1:qy])
            

            myPatch.fillBC()


            flux_x = splitFluxes('x',
                                 myPatch.nx, myPatch.ny, \
                                 myPatch.ng, \
                                 myPatch.dt, myPatch.x, myPatch.y, \
                                 dens)

            myPatch.data[0,0:qx-1,:] = myPatch.data[0,0:qx-1,:] + \
                                       dtdx*(flux_x[0:qx-1,:] - flux_x[1:qx,:])


        return 1

    else:
        print "ERROR: splitting " + splitting + " not supported."
        return -1
    
    
#------------------------------------------------------------------------------
# compute the advective timestep
#------------------------------------------------------------------------------

def timestep(myPatch):
    """ compute the CFL timestep for the current patch """

    cfl = runparams.getParam("driver.cfl")
    
    # get the variables we need
    dens = myPatch.getVar("density")

    # we need to compute the pressure
    u = runparams.getParam("advection.u")
    v = runparams.getParam("advection.v")
    
    # the timestep is min(dx/|u|, dy|v|)
    xtmp = myPatch.dx/max(abs(u),SMALL)
    ytmp = myPatch.dy/max(abs(v),SMALL)

    # there has to be an easier way to do this
    myPatch.dt = cfl*min(xtmp, ytmp)


