/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pyro.h"
#include "interface.h"


/* the HLLC Riemann solver, inputs are

    riemann.HLLC(1, nx, ny, ng, gamma, 
                 r_l, px_l, py_l, E_l,
		 r_r, px_r, py_r, E_r,
		 F_r, F_px, F_py, F_E)
 
   where idir is the direction (x = 1, y = 2) that we are jumping
   across
 
   gamma is the ratio of specific heats
 
   r_l, px_l, py_l, and E_l are the left conserved variable states
   r_r, px_r, py_r, and E_r are the left conserved variable states
 
   F_r, F_px, F_py, and F_E are the fluxes through the interface

*/


int _HLLC(int idir, int nx, int ny, int ng,
	  double gamma,
	  double **r_l, double **px_l, double **py_l, double **E_l,
	  double **r_r, double **px_r, double **py_r, double **E_r,
	  double **F_r, double **F_px, double **F_py, double **F_E) {

  /* solve the Riemann problem:
 
       Determine which of the possible wave states (rarefaction,
       contact, or shock) is present on the interface of two
       control volumes.  Take as input the left and right states
       and the polytropic gamma.  Return the HLLC flux on the interface
       and the state at the interface.
       
       The flux is computed using the states at the
       interface.  Corrections to the fluxes (i.e. HLLC solver)
       are done if we are in the star region.
 
 
        l_1      t    l_2       l_3
         \       ^   .       /
          \  *L  |   . *R   /
           \     |  .     /
            \    |  .    /
        L    \   | .   /    R
              \  | .  /
               \ |. /
                \|./
       ----------+----------------> x
 
       l_1 = u - cs   eigenvalue
       l_2 = u        eigenvalue (contact)
       l_3 = u + cs   eigenvalue
 
       only density jumps across l_2
 
 
       Note: we do not currently treat the transonic rarefaction
       properly.
 
  */
    
  int imin, imax, jmin, jmax;

  int i, j, ii;

  int ierr;

  double rho_avg, cs_avg;
  double p_star, u_star, rho_star_left, rho_star_right;
  double gamma_fac;
  double gammaFactor;
  
  double gammaMinus1, gammaPlus1, gammaInv;
  
  double Flux_l[NVAR], Flux_r[NVAR], U_l[NVAR], U_r[NVAR], U_state[NVAR];
  double Flux[NVAR];

  double cs_l, cs_r, cs_l_inv, cs_r_inv;
  double p_l, p_r, rho_l, rho_r, u_l, u_r, v_l, v_r, un_l, un_r, e_l, e_r;
  double p_min, p_max, Q;
  
  double A_l, A_r, B_l, B_r;
  double p_guess, p_lr;
  double g_l, g_r, pl_factor, pr_factor;
  double z;
  
  double S_l, S_c, S_r;
  
  double factor, factor2;
  
  waveType_t waveLeft, waveRight;
  waveRegion_t region;
  
  double HLLCfactor;
  

  imin = ng;
  imax = ng+nx-1;
 
  jmin = ng;
  jmax = ng+ny-1;
 
  /* loop over all the interfaces */

  for (i = imin-1; i <= imax+1; i++) {
    for (j = jmin-1; j <= jmax+1; j++) {

      /* primitive variables */
      rho_l = r_l[i][j];
      rho_r = r_r[i][j];

      u_l = px_l[i][j]/rho_l;
      u_r = px_r[i][j]/rho_r;

      v_l = py_l[i][j]/rho_l;
      v_r = py_r[i][j]/rho_r;

      p_l = (E_l[i][j] - 0.5*rho_l*(u_l*u_l + v_l*v_l))*(gamma - 1.0);
      p_r = (E_r[i][j] - 0.5*rho_r*(u_r*u_r + v_r*v_r))*(gamma - 1.0);

      e_l = E_l[i][j];
      e_r = E_r[i][j];
	

      /* input vectors, for convinence */
      U_l[DENS] = r_l[i][j];
      U_r[DENS] = r_r[i][j];
      
      U_l[XMOM] = px_l[i][j];
      U_r[XMOM] = px_r[i][j];

      U_l[YMOM] = py_l[i][j];
      U_r[YMOM] = py_r[i][j];

      U_l[ENER] = E_l[i][j];
      U_r[ENER] = E_r[i][j];


      /* Set the normal velocity -- if we are jumping in the x-direction,
	 then the x-velocity is the relevant one, otherwise, it is the
	 y-velocity */
      switch (idir) {
 
      case 1:
	un_l = u_l;
	un_r = u_r;
	break;
 
      case 2:
	un_l = v_l;
	un_r = v_r;
	break;
 
      default:
	printf("ERROR: invalid direction\n");
	exit(-1);
      }
      
      
      /* compute the sound speeds */
      cs_l = sqrt(gamma*p_l/rho_l);
      cs_r = sqrt(gamma*p_r/rho_r);
 
      cs_l_inv = 1.0/cs_l;
      cs_r_inv = 1.0/cs_r;
 
      /* compute some common gamma expressions */
      gammaMinus1 = gamma - 1.0;
      gammaPlus1  = gamma + 1.0;
 
      gammaInv = 1.0/gamma;
 
      
      /* ------------------------------------------------------------------
	 i. Estimate the star quantities -- use one of three methods to do
	 this -- the primitive variable Riemann solver, the two shock
	 approximation, or the two rarefaction approximation.  Pick the
	 method based on the pressure states at the interface.
	 ------------------------------------------------------------------ */
 
      p_max = max(p_l, p_r);
      p_min = min(p_l, p_r);
   
      Q = p_max/p_min;
 
      rho_avg = 0.5*(rho_l + rho_r);
      cs_avg = 0.5*(cs_l + cs_r);
     

      /* primitive variable Riemann solver (Toro, 9.3) */
 
      factor = rho_avg*cs_avg;
      factor2 = rho_avg/cs_avg;
   
      p_star = 0.5*(p_l + p_r) + 0.5*(un_l - un_r)*factor;
      u_star = 0.5*(un_l + un_r) + 0.5*(p_l - p_r)/factor;
   
      rho_star_left  = rho_l + (un_l - u_star)*factor2;
      rho_star_right = rho_r + (u_star - un_r)*factor2;
      
      if (Q > 2 && (p_star < p_min || p_star > p_max)) {
 
	/* use a more accurate Riemann solver for the estimate here */
 
	if (p_star < p_min) {
 
	  /* two rarefaction Riemann solver */
 
	  z = gammaMinus1/(2.0*gamma);
	  p_lr = pow((p_l/p_r), z);
 
	  u_star = (p_lr*un_l*cs_l_inv + un_r*cs_r_inv +
		    2.0*(p_lr - 1.0)/gammaMinus1)/
	    (p_lr*cs_l_inv + cs_r_inv);
 
	  pl_factor = 1.0 + 0.5*cs_l_inv*gammaMinus1*(un_l - u_star);
	  pr_factor = 1.0 + 0.5*cs_r_inv*gammaMinus1*(u_star - un_r);
 
	  p_star = 0.5*(p_l*pow(pl_factor, 1.0/z) +
			p_r*pow(pr_factor, 1.0/z));
 
	  rho_star_left  = rho_l*pow( (p_star/p_l), gammaInv );
	  rho_star_right = rho_r*pow( (p_star/p_r), gammaInv );
 
	} else {
 
	  /* two shock Riemann solver */
 
	  gammaFactor = gammaMinus1/gammaPlus1;
 
	  A_r = 2.0/(gammaPlus1 * rho_r);
	  B_r = gammaFactor*p_r;
 
	  A_l = 2.0/(gammaPlus1 * rho_l);
	  B_l = gammaFactor*p_l;
 
	  /* make a guess of the pressure using the PVRS above */
	  p_guess = max(0, p_star);
 
	  g_l = sqrt(A_l / (p_guess + B_l));
	  g_r = sqrt(A_r / (p_guess + B_r));
 
	  p_star = (g_l*p_l + g_r*p_r - (un_r - un_l))/(g_l + g_r);
 
	  u_star = 0.5*(un_l + un_r) +
	    0.5*( (p_star - p_r)*g_r - (p_star - p_l)*g_l);
           
	  rho_star_left = rho_l*((p_star/p_l) + gammaFactor)/
	    (gammaFactor*(p_star/p_l) + 1.0);
 
	  rho_star_right = rho_r*((p_star/p_r) + gammaFactor)/
	    (gammaFactor*(p_star/p_r) + 1.0);
 
	}
      }


      /* ------------------------------------------------------------------
	 ii. estimate the nonlinear wave speeds
	 ------------------------------------------------------------------ */
      
      /* determine what the non-linear waves are (rarefaction or shock) */
      if (p_star <= p_l)
	waveLeft = RAREFACTION;
      else
	waveLeft = SHOCK;
      
      
      if (p_star <= p_r)
	waveRight = RAREFACTION;
      else
	waveRight = SHOCK;
      
      
      /* compute the non-linear wave speeds */
      gamma_fac = (gamma + 1.0)/(2.0*gamma);
      
      if (waveLeft == RAREFACTION)
	S_l = un_l - cs_l;
      else {
	factor = 1.0 + gamma_fac*(p_star/p_l - 1.0);
	S_l = un_l - cs_l*pow(factor,0.5);
      }
      
      if (waveRight == RAREFACTION)
	S_r = un_r + cs_r;
      else {
	factor = 1.0 + gamma_fac*(p_star/p_r - 1.0);
	S_r = un_r + cs_r*pow(factor,0.5);
      }
      
      /*
	We could just take S_c = u_star as the estimate for the contact speed,
	but we can actually do this more accurately by using the Rankine-
	Hugonoit jump conditions across each of the waves (see Toro 10.58,
	Batten et al. SIAM J. Sci. and Stat. Comp., 18:1553 (1997)
      */
      
      S_c = (p_r - p_l + rho_l*un_l*(S_l - un_l) - rho_r*un_r*(S_r - un_r))/
	(rho_l*(S_l - un_l) - rho_r*(S_r - un_r));
      
 
      /* ------------------------------------------------------------------
	 iii. figure out which region we are in
	 ------------------------------------------------------------------ */
 
      if (S_r <= 0.0)
	region = RIGHT;
      else if (S_r > 0.0 && S_c <= 0.0)
	region = STAR_RIGHT;
      else if (S_c > 0.0 && S_l < 0.0)
	region = STAR_LEFT;
      else
	region = LEFT;
 
 
      /* ------------------------------------------------------------------
	 iv. now compute the state at the interface and the fluxes using
         the HLLC Riemann solver (Toro 10.4)
	 ------------------------------------------------------------------ */
 
      switch ( (int) region) {
      case RIGHT:
 
	for (ii=0; ii < NVAR; ii++) {
	  U_state[ii] = U_r[ii];
	}
 
	switch (idir) {
       
	case 1:
	  ierr = consFluxX(gamma, U_state, Flux);
	  break;
 
	case 2:
	  ierr = consFluxY(gamma, U_state, Flux);
	  break;
 
	}
 
	break;
 
 
      case STAR_RIGHT:
 
	HLLCfactor = rho_r*(S_r - un_r)/(S_r - S_c);
 
	U_state[DENS] = HLLCfactor;
 
	switch (idir) {
 
	case 1:
	  U_state[XMOM]   = HLLCfactor*S_c;
	  U_state[YMOM]   = HLLCfactor*v_r;
	  break;
 
	case 2:
	  U_state[XMOM]   = HLLCfactor*u_r;
	  U_state[YMOM]   = HLLCfactor*S_c;
	  break;
 
	}
       
	/* remember -- we are dealing with energy / volume NOT specific energy */
	U_state[ENER] = HLLCfactor*(e_r/rho_r +
				    (S_c - un_r)*
				    (S_c + p_r/(rho_r*(S_r - un_r))));
     
     
	/* find the flux on the right interface */
	switch (idir) {
 
	case 1:
	  ierr = consFluxX(gamma, U_r, Flux_r);
	  break;
 
	case 2:
	  ierr = consFluxY(gamma, U_r, Flux_r);
	  break;
 
	}
 
	for (ii = 0; ii < NVAR; ii++) {
	  Flux[ii] = Flux_r[ii] + S_r*(U_state[ii] - U_r[ii]);
	}
 
	break;
 
 
      case STAR_LEFT:
 
	HLLCfactor = rho_l*(S_l - un_l)/(S_l - S_c);
 
	U_state[DENS] = HLLCfactor;
 
	switch (idir) {
 
	case 1:
	  U_state[XMOM] = HLLCfactor*S_c;
	  U_state[YMOM] = HLLCfactor*v_l;
	  break;
 
	case 2:
	  U_state[XMOM] = HLLCfactor*u_l;
	  U_state[YMOM] = HLLCfactor*S_c;
	  break;
 
	}
 
	/* remember -- we are dealing with energy / volume NOT specific energy */
	U_state[ENER] = HLLCfactor*(e_l/rho_l +
				    (S_c - un_l)*
				    (S_c + p_l/(rho_l*(S_l - un_l))));
     
     
	/* find the flux on the right interface */
	switch (idir) {
 
	case 1:
	  ierr = consFluxX(gamma, U_l, Flux_l);
	  break;
 
	case 2:
	  ierr = consFluxY(gamma, U_l, Flux_l);
	  break;
 
	}
 
	for (ii = 0; ii < NVAR; ii++) {
	  Flux[ii] = Flux_l[ii] + S_l*(U_state[ii] - U_l[ii]);
	}
 
	break;
 
 
      case LEFT:
 
	for (ii=0; ii < NVAR; ii++) {
	  U_state[ii] = U_l[ii];
	}
 
	switch (idir) {
	  
	case 1:
	  ierr = consFluxX(gamma, U_state, Flux);
	  break;
 
	case 2:
	  ierr = consFluxY(gamma, U_state, Flux);
	  break;
 
	}
 
	break;
 
      default:
	printf("error region not defined in Riemann solver\n");
	exit(-1);
      }
 
 
  
      F_r[i][j]  = Flux[DENS];
      F_px[i][j] = Flux[XMOM];
      F_py[i][j] = Flux[YMOM];
      F_E[i][j]  = Flux[ENER];
 
    }
  }

      
  return 1;

}

