/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale
   
  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
 
#include "Python.h"
#include "numarray/numarray.h"
#include "numarray/libnumarray.h"

#include "pyro.h"
#include "pyroarray.h"
#include "interface.h"   

/* wrappers to pass information back and forth between python and C */


static PyObject *unsplitFluxes(PyObject *self, PyObject *args) {

  PyObject *odens, *oxmom, *oymom, *oener, *ospecies, *oF_x, *oF_y;
  PyArrayObject *pdens, *pxmom, *pymom, *pener, *pspecies, *pF_x, *pF_y;

  double **dens, **xmom, **ymom, **ener, ***species, ***F_x, ***F_y;

  double dt, dx, dy, cvisc, gamma, grav;

  int qx, qy;

  int ierr;

  int nx, ny, ng, nspecies;

  /* parse the inputs:

     _unsplitFluxes(nx, ny, ng, nspecies,
                    dt, dx, dy, 
		    cvisc,
		    gamma, grav,
		    dens, xmom, ymom, ener, species,
		    F_x, F_y)

     nx/ny = number of zones in x/y (interior)
     ng = number of guard cells
     nspecies = number of species

     dt = timestep
     dx/dx = x/y grid spacing

     cvisc = artifical viscosity coefficient

     gamma = gamma
     grav = gravitational acceleration

     dens = density
     xmom = x-momentum
     ymom = y-momentum
     ener = energy
     species = advected species

     F_x/F_y = flux through the x/y-interface

  */
  
  if (!PyArg_ParseTuple(args, "iiiiddddddOOOOOOO",
			&nx, &ny, &ng, &nspecies,
			&dt, &dx, &dy,
			&cvisc,
			&gamma, &grav, 
			&odens, &oxmom, &oymom, &oener, &ospecies,
			&oF_x, &oF_y))
    return NULL;

  qx = 2*ng + nx;
  qy = 2*ng + ny;

  /* transform the PyObject to a C array using the pyro py2C routines */
  py2C_init();
  dens = py2C_create2d(odens, qx, qy, &pdens);
  xmom = py2C_create2d(oxmom, qx, qy, &pxmom);
  ymom = py2C_create2d(oymom, qx, qy, &pymom);
  ener = py2C_create2d(oener, qx, qy, &pener);
  species = py2C_create3d(ospecies, nspecies, qx, qy, &pspecies);
  F_x = py2C_create3d(oF_x, NVAR+nspecies, qx, qy, &pF_x);
  F_y = py2C_create3d(oF_y, NVAR+nspecies, qx, qy, &pF_y);

  /* actually call the states routine now */
  ierr = _unsplitFluxes(nx, ny, ng, nspecies,
		        dt, dx, dy,
			cvisc,
		        gamma, grav, 
		        dens, xmom, ymom, ener,
		        species,
		        F_x, F_y);
  if (ierr < 0) {
    printf("ERROR in unsplitFluxes: illegal return value from _unsplitFluxes\n");
    return NULL;
  }

  /* clean-up */
  py2C_free2d(dens, qx, qy, pdens);
  py2C_free2d(xmom, qx, qy, pxmom);
  py2C_free2d(ymom, qx, qy, pymom);
  py2C_free2d(ener, qx, qy, pener);
  py2C_free3d(species, nspecies, qx, qy, pspecies);
  py2C_free3d(F_x, NVAR+nspecies, qx, qy, pF_x);
  py2C_free3d(F_y, NVAR+nspecies, qx, qy, pF_y);

  return Py_BuildValue("i", 1);

}


static PyObject *splitFluxes(PyObject *self, PyObject *args) {

  PyObject *odens, *oxmom, *oymom, *oener, *ospecies, *oF_x;
  PyArrayObject *pdens, *pxmom, *pymom, *pener, *pspecies, *pF_x;

  double **dens, **xmom, **ymom, **ener, ***species, ***F_x;

  double dt, dx, dy, gamma, grav;

  int qx, qy;

  int ierr;

  int idir, nx, ny, ng, nspecies;

  /* parse the inputs:

     _splitFluxes(idir, nx, ny, ng, nspecies,
                  dt, dx, dy, 
		  gamma, grav,
		  dens, xmom, ymom, ener, species,
		  F_x)

     idir = direction to do the flux computation
     nx/ny = number of zones in x/y (interior)
     ng = number of guard cells
     nspecies = number of species

     dt = timestep
     dx/dy = x/y grid spacing

     gamma = gamma
     grav = gravitational acceleration

     dens = density
     xmom = x-momentum
     ymom = y-momentum
     ener = energy
     species = advected species

     F_x = flux through the interface

  */
  
  if (!PyArg_ParseTuple(args, "iiiiidddddOOOOOO",
			&idir, &nx, &ny, &ng, &nspecies,
			&dt, &dx, &dy,
			&gamma, &grav, 
			&odens, &oxmom, &oymom, &oener, &ospecies,
			&oF_x))
    return NULL;

  qx = 2*ng + nx;
  qy = 2*ng + ny;

  /* transform the PyObject to a C array using the pyro py2C routines */
  py2C_init();
  dens = py2C_create2d(odens, qx, qy, &pdens);
  xmom = py2C_create2d(oxmom, qx, qy, &pxmom);
  ymom = py2C_create2d(oymom, qx, qy, &pymom);
  ener = py2C_create2d(oener, qx, qy, &pener);
  species = py2C_create3d(ospecies, nspecies, qx, qy, &pspecies);
  F_x = py2C_create3d(oF_x, NVAR+nspecies, qx, qy, &pF_x);

  /* actually call the splitFluxes routine now */
  ierr = _splitFluxes(idir, nx, ny, ng, nspecies,
		      dt, dx, dy,
		      gamma, grav, 
		      dens, xmom, ymom, ener,
		      species,
		      F_x);
  if (ierr < 0) {
    printf("ERROR in splitFluxes: illegal return value from _splitFluxes\n");
    return NULL;
  }

  /* clean-up */
  py2C_free2d(dens, qx, qy, pdens);
  py2C_free2d(xmom, qx, qy, pxmom);
  py2C_free2d(ymom, qx, qy, pymom);
  py2C_free2d(ener, qx, qy, pener);
  py2C_free3d(species, nspecies, qx, qy, pspecies);
  py2C_free3d(F_x, NVAR+nspecies, qx, qy, pF_x);

  return Py_BuildValue("i", 1);

}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyMethodDef compressiblemethods[] = {
  {"unsplitFluxes", unsplitFluxes, METH_VARARGS},
  {"splitFluxes", splitFluxes, METH_VARARGS},
  {NULL, NULL}
};
 
void initcompressible_fluxes(void) {
  Py_InitModule("compressible_fluxes", compressiblemethods);
  import_libnumarray();
}
