"""
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
"""
import eos
from util import runparams
from mesh import *
from numarray import *
from compressible_fluxes import *

def initialize(myPatch):
    """ initialize the variables for compressible hydro """

    # start by making sure that grid is a valid grid object
    if not isinstance(myPatch, patch.cellCentered):
        print "ERROR: patch invalid in io.dumpPatch()"
        print myPatch.__class__
        return -1                          

    myPatch.registerVar("density")
    myPatch.registerVar("x-momentum")
    myPatch.registerVar("y-momentum")
    myPatch.registerVar("energy")


def evolve(myPatch):
    """ evolve the compressible equations through one timestep """

    dens = myPatch.getVar("density")
    xmom = myPatch.getVar("x-momentum")
    ymom = myPatch.getVar("y-momentum")
    ener = myPatch.getVar("energy")    

    splitting = runparams.getParam("driver.splitting")

    qx = 2*myPatch.ng + myPatch.nx
    qy = 2*myPatch.ng + myPatch.ny

    dtdx = myPatch.dt/myPatch.dx
    dtdy = myPatch.dt/myPatch.dy

    grav = runparams.getParam("compressible.grav")
    gamma = runparams.getParam("eos.gamma")    

    cvisc = runparams.getParam("compressible.cvisc")
    
    if splitting == "unsplit":
        
        nspecies = 1
        species = zeros((nspecies, qx, qy), Float64)

        flux_x = zeros((myPatch.nvar + nspecies, qx, qy), Float64)
        flux_y = zeros((myPatch.nvar + nspecies, qx, qy), Float64)        


        unsplitFluxes(myPatch.nx, myPatch.ny, myPatch.ng, nspecies,
                      myPatch.dt, myPatch.dx, myPatch.dy,
                      cvisc,
                      gamma, grav,
                      dens, xmom, ymom, ener, species,
                      flux_x, flux_y)
                      
        
        old_dens = myPatch.data[0,:,:].copy()
        old_momy = myPatch.data[2,:,:].copy()

        myPatch.data[:,0:qx-1,0:qy-1] += \
              dtdx*(flux_x[0:myPatch.nvar,0:qx-1,0:qy-1] - \
                    flux_x[0:myPatch.nvar,1:qx,0:qy-1]) + \
              dtdy*(flux_y[0:myPatch.nvar,0:qx-1,0:qy-1] - \
                    flux_y[0:myPatch.nvar,0:qx-1,1:qy])

        # add the source terms
        myPatch.data[2,:,:] += 0.5*myPatch.dt*(myPatch.data[0,:,:] + old_dens[:,:])*grav
        myPatch.data[3,:,:] += 0.5*myPatch.dt*(myPatch.data[2,:,:] + old_momy[:,:])*grav
        
        return 1

        
    elif splitting == "strang" or splitting == "Strang":

        # when we Strang-split, we want to do X(dt) Y(dt) Y(dt) X(dt) U,
        # to advance U two timesteps.  We do X Y in even steps, and
        # Y X in odd steps to get the right symmetry.
        
        if myPatch.nstep % 2 == 0:

            # XY
            nspecies = 1
            species = zeros((nspecies, qx, qy), Float64)

            flux_x = zeros((myPatch.nvar + nspecies, qx, qy), Float64)
            flux_y = zeros((myPatch.nvar + nspecies, qx, qy), Float64)        


            splitFluxes(1, myPatch.nx, myPatch.ny, myPatch.ng, nspecies,
                        myPatch.dt, myPatch.dx, myPatch.dy,
                        gamma, grav,
                        dens, xmom, ymom, ener, species,
                        flux_x)
                      
            myPatch.data[:,0:qx-1,:] += \
                  dtdx*(flux_x[0:myPatch.nvar,0:qx-1,:] - \
                        flux_x[0:myPatch.nvar:,1:qx,:])


            myPatch.fillBC()


            splitFluxes(2, myPatch.nx, myPatch.ny, myPatch.ng, nspecies,
                        myPatch.dt, myPatch.dx, myPatch.dy,
                        gamma, grav,
                        dens, xmom, ymom, ener, species,
                        flux_y)
                                  
            old_dens = myPatch.data[0,:,:].copy()
            old_momy = myPatch.data[2,:,:].copy()

            myPatch.data[:,:,0:qy-1] += \
                  dtdy*(flux_y[0:myPatch.nvar,:,0:qy-1] - \
                        flux_y[0:myPatch.nvar,:,1:qy])

            # add the y source terms
            myPatch.data[2,:,:] += 0.5*myPatch.dt*(myPatch.data[0,:,:] + old_dens[:,:])*grav
            myPatch.data[3,:,:] += 0.5*myPatch.dt*(myPatch.data[2,:,:] + old_momy[:,:])*grav
            
        else:

            # YX
            nspecies = 1
            species = zeros((nspecies, qx, qy), Float64)

            flux_x = zeros((myPatch.nvar + nspecies, qx, qy), Float64)
            flux_y = zeros((myPatch.nvar + nspecies, qx, qy), Float64)
            
            splitFluxes(2, myPatch.nx, myPatch.ny, myPatch.ng, nspecies,
                        myPatch.dt, myPatch.dx, myPatch.dy,
                        gamma, grav,
                        dens, xmom, ymom, ener, species,
                        flux_y)
            
            old_dens = myPatch.data[0,:,:].copy()
            old_momy = myPatch.data[2,:,:].copy()

            myPatch.data[:,:,0:qy-1] += \
                  dtdy*(flux_y[0:myPatch.nvar,:,0:qy-1] - \
                        flux_y[0:myPatch.nvar,:,1:qy])            

            # add the y source terms
            myPatch.data[2,:,:] += 0.5*myPatch.dt*(myPatch.data[0,:,:] + old_dens[:,:])*grav
            myPatch.data[3,:,:] += 0.5*myPatch.dt*(myPatch.data[2,:,:] + old_momy[:,:])*grav

            myPatch.fillBC()


            splitFluxes(1, myPatch.nx, myPatch.ny, myPatch.ng, nspecies,
                        myPatch.dt, myPatch.dx, myPatch.dy,
                        gamma, grav,
                        dens, xmom, ymom, ener, species,
                        flux_x)

            myPatch.data[:,0:qx-1,:] += \
                  dtdx*(flux_x[0:myPatch.nvar,0:qx-1,:] - \
                        flux_x[0:myPatch.nvar,1:qx,:])
            
        return 1

    
    else:
        print "ERROR: splitting " + splitting + " no supported"
        return -1

    
#------------------------------------------------------------------------------
# compute the compressible timestep
#------------------------------------------------------------------------------

def timestep(myPatch):
    """ compute the CFL timestep for the current patch """

    cfl = runparams.getParam("driver.cfl")
    
    # get the variables we need
    dens = myPatch.getVar("density")
    xmom = myPatch.getVar("x-momentum")
    ymom = myPatch.getVar("y-momentum")
    ener = myPatch.getVar("energy")

    idens = 1.0/dens

    
    # we need to compute the pressure
    u = xmom*idens
    v = ymom*idens
    
    eint = (ener - 0.5*dens*(u*u + v*v))*idens

    p = eos.pres(dens, eint)

    # compute the sounds speed
    gamma = runparams.getParam("eos.gamma")

    cs = sqrt(gamma*p*idens)
    
    # the timestep is min(dx/(|u| + cs), dy/(|v| + cs))
    xtmp = myPatch.dx/(abs(u) + cs)
    ytmp = myPatch.dy/(abs(v) + cs)

    myPatch.dt = cfl*minimum(xtmp.min(), ytmp.min())
