/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale
 
  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pyro.h"
#include "interface.h"


int _states(int idir, int nx, int ny, int ng, int nspecies,
	    double gamma, double dtdx,
	    double **r, double **u, double **v, double **p, double ***xn,
	    double **ldr, double **ldu, double **ldv, double **ldp, double ***ldxn,
	    double **r_xl, double **r_xr, 
	    double **u_xl, double **u_xr,
	    double **v_xl, double **v_xr,
	    double **p_xl, double **p_xr,
	    double ***xn_xl, double ***xn_xr) {


  double dtdx4;
  
  double *dV;
  double *eval;

  int i, j, m, n;

  int imin, imax, jmin, jmax;

  double cs;

  /* eigenvectors.  The first index is the vector index, the second 
     corresponds to the components */
  double **lvec, *lvec_data, **rvec, *rvec_data;

  double *betal, *betar;

  double sum, sum_l, sum_r;


  /* there are 4+nspecies eigenvalues, eigenvectors, etc. */
  dV = (double *) malloc(sizeof(double)*(NVAR+nspecies));
  eval = (double *) malloc(sizeof(double)*(NVAR+nspecies));

  lvec_data = (double *) malloc(sizeof(double)*(NVAR+nspecies)*(NVAR+nspecies));
  lvec = cArray_2d(lvec_data, NVAR+nspecies, NVAR+nspecies);

  rvec_data = (double *) malloc(sizeof(double)*(NVAR+nspecies)*(NVAR+nspecies));
  rvec = cArray_2d(rvec_data, NVAR+nspecies, NVAR+nspecies);
  
  betal = (double *) malloc(sizeof(double)*(NVAR+nspecies));
  betar = (double *) malloc(sizeof(double)*(NVAR+nspecies));
  


  imin = ng;
  imax = ng+nx-1;

  jmin = ng;
  jmax = ng+ny-1;

  /* We need the left and right eigenvectors and the eigenvalues for the
     system projected along the x-direction
 
     The eigenvalues are u - c, u, u, u + c, u, u, ... u

     There are 2 + nspecies u eigenvalues.

 
     The Jacobian matrix for the primitive variable formulation of the
     Euler equations projected in the x-direction is:
     
           / u   r   0   0  ...      0  ...  0 \ 
           | 0   u   0  1/r ...      0  ...  0 |
       A = | 0   0   u   0  ...      0  ...  0 |
           | 0  rc^2 0   u  ...      0  ...  0 |
           | 0   0   0   0   .       0       0 |
           | :   :   :   :       .   :       0 |
           | 0   0   0   0  ...      u  ...  0 |
           | :   :   :   :               .   0 |
           \ 0   0   0   0  ...      0  ...  u /

     A is a NVAR+nspecies x NVAR+nspecies array.  The first NVAR (=4) rows 
     correspond to rho, u, v, p.  Each species X_k has a u on the
     diagonal, and zeros everywhere else in 4+kth row/column.
     
     The right eigenvectors are
     
           /  1  \        / 1 \        / 0 \        /  1  \               / 0 \
           |-c/r |        | 0 |        | 0 |        | c/r |               | 0 |
      r1 = |  0  |   r2 = | 0 |   r3 = | 1 |   r4 = |  0  |  ... r(4+k) = | 0 |
           | c^2 |        | 0 |        | 0 |        | c^2 |               | 0 |
           |  0  |        | 0 |        | 0 |        |  0  |               | 0 |
           |  :  |        | : |        | : |        |  :  |               | : |
           |  :  |        | : |        | : |        |  :  |               | 1 |
           |  :  |        | : |        | : |        |  :  |               | : |
           \  0  /        \ 0 /        \ 0 /        \  0  /               \ 0 /

     After the first 4 rows, all entries are 0 except for the 4+kth row in the
     4+kth eigenvector, which is 1 -- i.e. each species is simply advected.

     
     The left eigenvectors are

       l1 =     ( 0,  -r/(2c),  0, 1/(2c^2),  0, ... 0, ... 0 )
     
       l2 =     ( 1,     0,     0,  -1/c^2,   0, ... 0, ... 0 )
     
       l3 =     ( 0,     0,     1,     0,     0, ... 0, ... 0 )
     
       l4 =     ( 0,   r/(2c),  0,  1/(2c^2), 0, ... 0, ... 0 )
  
       :

       l(4+k) = ( 0,     0,     0,     0,     0, ... 1, ... 0 )

     After the first 4 eigenvectors, all entries are 0 except for the 4+kth
     entry in the 4+kth eigenvector, which is 1.


     the fluxes are going to be defined on the left edge of the
     computational zones
 
 
              |             |             |             |
              |             |             |             |
             -+------+------+------+------+------+------+--
              |     i-1     |      i      |     i+1     |
 
                      V_l,i  V_r,i   V_l,i+1
 
 
     V_r,i and V_l,i+1 are computed using the information in
     zone i,j.
 
 
  */

  dtdx4 = 0.25*dtdx;

  for (i = imin-2; i <= imax+2; i++) {
    for (j = jmin-2; j <= jmax+2; j++) {

      dV[DENS] = ldr[i][j];
      dV[XVEL] = ldu[i][j];
      dV[YVEL] = ldv[i][j];
      dV[PRES] = ldp[i][j];

      for (n = 0; n < nspecies; n++) {
	dV[NVAR+n] = ldxn[n][i][j];
      }

      cs = sqrt(gamma*p[i][j]/r[i][j]);

      if (idir == 1) {

	/* compute the eigenvalues */
	eval[0] = u[i][j] - cs;
	eval[1] = u[i][j];
	eval[2] = u[i][j];
	eval[3] = u[i][j] + cs;
	
	for (n = 0; n < nspecies; n++) {
	  eval[NVAR+n] = u[i][j];
	}


	zero_2d(lvec, NVAR+nspecies, NVAR+nspecies);

	/* compute left eigenvectors */
	/* u - c eigenvector */
	lvec[0][0] = 0.0;
	lvec[0][1] = -0.5*r[i][j]/cs;
	lvec[0][2] = 0.0;
	lvec[0][3] = 0.5/(cs*cs);
	/* lvec[0][NVAR...] = 0 */
	
	/* u eigenvector */
	lvec[1][0] = 1.0;
	lvec[1][1] = 0.0;
	lvec[1][2] = 0.0;
	lvec[1][3] = -1.0/(cs*cs);
	/* lvec[1][NVAR...] = 0 */

	/* u eigenvector */
	lvec[2][0] = 0.0;
	lvec[2][1] = 0.0;
	lvec[2][2] = 1.0;
	lvec[2][3] = 0.0;
	/* lvec[2][NVAR...] = 0 */
	
	/* u + c eigenvector */
	lvec[3][0] = 0.0;
	lvec[3][1] = 0.5*r[i][j]/cs;
	lvec[3][2] = 0.0;
	lvec[3][3] = 0.5/(cs*cs);
	/* lvec[3][NVAR...] = 0 */	

	/* u eigenvectors corresponding to the species */
	for (n = 0; n < nspecies; n++) {
	  lvec[NVAR+n][NVAR+n] = 1.0;
	}
	

	zero_2d(rvec, NVAR+nspecies, NVAR+nspecies);

	/* compute right eigenvectors */
	/* u - c eigenvector */
	rvec[0][0] = 1.0;
	rvec[0][1] = -cs/r[i][j];
	rvec[0][2] = 0.0;
	rvec[0][3] = cs*cs;
	/* rvec[0][NVAR...] = 0 */

	/* u eigenvector */
	rvec[1][0] = 1.0;
	rvec[1][1] = 0.0;
	rvec[1][2] = 0.0;
	rvec[1][3] = 0.0;
	/* rvec[1][NVAR...] = 0 */

	/* u eigenvector */
	rvec[2][0] = 0.0;
	rvec[2][1] = 0.0;
	rvec[2][2] = 1.0;
	rvec[2][3] = 0.0;
	/* rvec[2][NVAR...] = 0 */

	/* u + c eigenvector */
	rvec[3][0] = 1.0;
	rvec[3][1] = cs/r[i][j];
	rvec[3][2] = 0.0;
	rvec[3][3] = cs*cs;
	/* rvec[0][NVAR...] = 0 */

	/* u eigenvectors corresponding to the species */
	for (n = 0; n < nspecies; n++) {
	  rvec[NVAR+n][NVAR+n] = 1.0;
	}


      } else if (idir == 2) {

	/* compute the eigenvalues */
	eval[0] = v[i][j] - cs;
	eval[1] = v[i][j];
	eval[2] = v[i][j];
	eval[3] = v[i][j] + cs;

	for (n = 0; n < nspecies; n++) {
	  eval[NVAR+n] = v[i][j];
	}


	zero_2d(lvec, NVAR+nspecies, NVAR+nspecies);

	/* compute left eigenvectors */
	/* u - c eigenvector */
	lvec[0][0] = 0.0;
	lvec[0][1] = 0.0;
	lvec[0][2] = -0.5*r[i][j]/cs;
	lvec[0][3] = 0.5/(cs*cs);
	/* lvec[0][NVAR...] = 0 */
	
	/* u eigenvector */
	lvec[1][0] = 1.0;
	lvec[1][1] = 0.0;
	lvec[1][2] = 0.0;
	lvec[1][3] = -1.0/(cs*cs);
	/* lvec[1][NVAR...] = 0 */

	/* u eigenvector */
	lvec[2][0] = 0.0;
	lvec[2][1] = 1.0;
	lvec[2][2] = 0.0;
	lvec[2][3] = 0.0;
	/* lvec[2][NVAR...] = 0 */
	
	/* u + c eigenvector */
	lvec[3][0] = 0.0;
	lvec[3][1] = 0.0;
	lvec[3][2] = 0.5*r[i][j]/cs;
	lvec[3][3] = 0.5/(cs*cs);
	/* lvec[3][NVAR...] = 0 */

	/* u eigenvectors corresponding to the species */
	for (n = 0; n < nspecies; n++) {
	  lvec[NVAR+n][NVAR+n] = 1.0;
	}


	zero_2d(rvec, NVAR+nspecies, NVAR+nspecies);

	/* compute right eigenvectors */
	/* u - c eigenvector */
	rvec[0][0] = 1.0;
	rvec[0][1] = 0.0;
	rvec[0][2] = -cs/r[i][j];
	rvec[0][3] = cs*cs;
	/* rvec[0][NVAR...] = 0 */
	
	/* u eigenvector */
	rvec[1][0] = 1.0;
	rvec[1][1] = 0.0;
	rvec[1][2] = 0.0;
	rvec[1][3] = 0.0;
	/* rvec[1][NVAR...] = 0 */

	/* u eigenvector */
	rvec[2][0] = 0.0;
	rvec[2][1] = 1.0;
	rvec[2][2] = 0.0;
	rvec[2][3] = 0.0;
	/* rvec[2][NVAR...] = 0 */

	/* u + c eigenvector */
	rvec[3][0] = 1.0;
	rvec[3][1] = 0.0;
	rvec[3][2] = cs/r[i][j];
	rvec[3][3] = cs*cs;
	/* rvec[3][NVAR...] = 0 */

	/* u eigenvectors corresponding to the species */
	for (n = 0; n < nspecies; n++) {
	  rvec[NVAR+n][NVAR+n] = 1.0;
	}


      } else {
	printf("ERROR: direction not supported in states\n");
	return -1;
      }

      if (idir == 1) {

	/* define the reference states -- the maximum eigenvalue is 
	   eval[3] = u + c */
	r_xl[i+1][j] = r[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldr[i][j];
	r_xr[i][j] = r[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldr[i][j];
	
	u_xl[i+1][j] = u[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldu[i][j];
	u_xr[i][j] = u[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldu[i][j];
	
	v_xl[i+1][j] = v[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldv[i][j];
	v_xr[i][j] = v[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldv[i][j];

	p_xl[i+1][j] = p[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldp[i][j];
	p_xr[i][j] = p[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldp[i][j];

	for (n = 0; n < nspecies; n++) {
	  xn_xl[n][i+1][j] = xn[n][i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldxn[n][i][j];
	  xn_xr[n][i][j] = xn[n][i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldxn[n][i][j];
	}
	

      } else if (idir == 2) {

	/* define the reference states */
	r_xl[i][j+1] = r[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldr[i][j];
	r_xr[i][j] = r[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldr[i][j];
	
	u_xl[i][j+1] = u[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldu[i][j];
	u_xr[i][j] = u[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldu[i][j];
	
	v_xl[i][j+1] = v[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldv[i][j];
	v_xr[i][j] = v[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldv[i][j];
	
	p_xl[i][j+1] = p[i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldp[i][j];
	p_xr[i][j] = p[i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldp[i][j];

	for (n = 0; n < nspecies; n++) {
	  xn_xl[n][i][j+1] = xn[n][i][j] + 0.5*(1.0 - dtdx*max(eval[3], 0.0))*ldxn[n][i][j];
	  xn_xr[n][i][j] = xn[n][i][j] - 0.5*(1.0 + dtdx*min(eval[0], 0.0))*ldxn[n][i][j];
	}

      } else {
	printf("ERROR: direction not supported in states\n");
	return -1;
      }

      /* compute the Vhat functions here */
      for (m = 0; m < NVAR+nspecies; m++) {

	sum = 0.0;
	for (n = 0; n < NVAR+nspecies; n++) {
	  sum = sum + lvec[m][n]*dV[n];
	}
	
	betal[m] = dtdx4*(eval[3] - eval[m])*(sgn(eval[m]) + 1.0)*sum;
	
	betar[m] = dtdx4*(eval[0] - eval[m])*(1.0 - sgn(eval[m]))*sum;
      }


      /* density */
      sum_l = 0.0;
      sum_r = 0.0;
      for (n = 0; n < NVAR+nspecies; n++) {
	sum_l = sum_l + betal[n]*rvec[n][0];
	sum_r = sum_r + betar[n]*rvec[n][0];
      }

      if (idir == 1) {
	r_xl[i+1][j] = r_xl[i+1][j] + sum_l;
	r_xr[i][j] = r_xr[i][j] + sum_r;

      } else if (idir == 2) {
	r_xl[i][j+1] = r_xl[i][j+1] + sum_l;
	r_xr[i][j] = r_xr[i][j] + sum_r;
      }

      
      /* x-velocity */
      sum_l = 0.0;
      sum_r = 0.0;
      for (n = 0; n < NVAR+nspecies; n++) {
	sum_l = sum_l + betal[n]*rvec[n][1];
	sum_r = sum_r + betar[n]*rvec[n][1];
      }

      if (idir == 1) {
	u_xl[i+1][j] = u_xl[i+1][j] + sum_l;
	u_xr[i][j] = u_xr[i][j] + sum_r;

      } else if (idir == 2) {
	u_xl[i][j+1] = u_xl[i][j+1] + sum_l;
	u_xr[i][j] = u_xr[i][j] + sum_r;
      }


      /* y-velocity */
      sum_l = 0.0;
      sum_r = 0.0;
      for (n = 0; n < NVAR+nspecies; n++) {
	sum_l = sum_l + betal[n]*rvec[n][2];
	sum_r = sum_r + betar[n]*rvec[n][2];
      }

      if (idir == 1) {
	v_xl[i+1][j] = v_xl[i+1][j] + sum_l;
	v_xr[i][j] = v_xr[i][j] + sum_r;

      } else if (idir == 2) {
	v_xl[i][j+1] = v_xl[i][j+1] + sum_l;
	v_xr[i][j] = v_xr[i][j] + sum_r;
      }


      /* pressure */
      sum_l = 0.0;
      sum_r = 0.0;
      for (n = 0; n < NVAR+nspecies; n++) {
	sum_l = sum_l + betal[n]*rvec[n][3];
	sum_r = sum_r + betar[n]*rvec[n][3];
      }

      if (idir == 1) {
	p_xl[i+1][j] = p_xl[i+1][j] + sum_l;
	p_xr[i][j] = p_xr[i][j] + sum_r;

      } else if (idir == 2) {
	p_xl[i][j+1] = p_xl[i][j+1] + sum_l;
	p_xr[i][j] = p_xr[i][j] + sum_r;
      }


      /* species */
      for (m = 0; m < nspecies; m++) {
	  
	sum_l = 0.0;
	sum_r = 0.0;
	for (n = 0; n < NVAR+nspecies; n++) {
	  sum_l = sum_l + betal[n]*rvec[n][NVAR+m];
	  sum_r = sum_r + betar[n]*rvec[n][NVAR+m];
	}

	if (idir == 1) {
	  xn_xl[m][i+1][j] = xn_xl[m][i+1][j] + sum_l;
	  xn_xr[m][i][j] = xn_xr[m][i][j] + sum_r;
	} else if (idir == 2) {
	  xn_xl[m][i][j+1] = xn_xl[m][i][j+1] + sum_l;
	  xn_xr[m][i][j] = xn_xr[m][i][j] + sum_r;
	}
	  
      }
      
      
    }
  }

  free(dV);
  free(eval);

  free(betal);
  free(betar);

  free(lvec_data);
  free(rvec_data);

  free_2d(lvec);
  free_2d(rvec);

  return 1;
}

