#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pyro.h"
#include "interface.h"
#include "reconstruction.h"

/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------

Split implementation of the Colella 2nd order unsplit Godunov scheme.  This
is a 2-dimensional implementation only.  We assume that the grid is
uniform, but it is relatively straightforward to relax this
assumption.

There are several different options for this solver (they are all
discussed in the Colella paper).

  limiter          = 1 to use the 2nd order MC limiter
                   = 2 to use the 4th order MC limiter

  use_flattening   = t to use the multidimensional flattening
                     algorithm at shocks

  delta, z0, z1      these are the flattening parameters.  The default
                     are the values listed in Colella 1990.


   j+3/2--+---------+---------+---------+
          |         |         |         |
     j+1 _|         |         |         |
          |         |         |         |
          |         |         |         |
   j+1/2--+---------XXXXXXXXXXX---------+
          |         X         X         |
       j _|         X         X         |
          |         X         X         |
          |         X         X         |
   j-1/2--+---------XXXXXXXXXXX---------+
          |         |         |         |
     j-1 _|         |         |         |
          |         |         |         |
          |         |         |         |
   j-3/2--+---------+---------+---------+
          |    |    |    |    |    |    |    
              i-1        i        i+1        
        i-3/2     i-1/2     i+1/2     i+3/2   

We wish to solve

  U_t + F^x_x + F^y_y = H

we do this via dimensional splitting, first solving

  U_t + F^x_x = H

and then the y-projection.

we want U_{i+1/2}^{n+1/2} -- the interface values that are input to
the Riemann problem through the faces for each zone.

Taylor expanding yields

   n+1/2                     dU           dU  
  U          = U   + 0.5 dx  --  + 0.5 dt --
   i+1/2,j,L    i,j          dx           dt


                             dU             dF^x 
             = U   + 0.5 dx  --  - 0.5 dt ( ---- - H )
                i,j          dx              dx  


                              dU      dF^x   
             = U   + 0.5 ( dx -- - dt ---- ) + 0.5 dt H
                i,j           dx       dx    


                                  dt       dU 
             = U   + 0.5 dx ( 1 - -- A^x ) -- + 0.5 dt H
                i,j               dx       dx 


                                dt       _  
             = U   + 0.5  ( 1 - -- A^x ) DU + 0.5 dt H
                i,j             dx          


                     +----------+----------+ +---+---+
                                |                |
 
                    this is the monotonized   source term
                    central difference term  
                                             
This is the split version!
          
*/

/*===========================================================================
  splitFluxes
  =========================================================================== */

int _splitFluxes(int idir, int nx, int ny, int ng, int nspecies,
		double dt, double dx, double dy, 
		double gamma, double grav,
		double **dens, double **xmom, double **ymom, double **ener,
		double ***species,
		double ***F)
{
  
  /*
    splitFluxes returns the fluxes through the idir interfaces by
    doing a split reconstruction of the interface values and then
    solving the Riemann problem through all the interfaces at once
    
    currently we assume a gamma-law EOS
    
    grav is the gravitational acceleration in the y-direction
  */

  int i, j, n;

  int imin, imax;
  int jmin, jmax;
  int qx, qy;

  double **r;
  double ri;
  double **u, *u_data;
  double **v, *v_data;
  double **p, *p_data;
  double ***xn, *xn_data;

  double **xi, *xi_data;

  double ***U_l, *U_l_data, ***U_r, *U_r_data;

  double **ldelta_r, *ldelta_r_data;
  double **ldelta_u, *ldelta_u_data;
  double **ldelta_v, *ldelta_v_data;
  double **ldelta_p, *ldelta_p_data;
  double ***ldelta_xn, *ldelta_xn_data;

  double ***V_l, *V_l_data, ***V_r, *V_r_data;

  double delta, z0, z1, smallp;

  double dtdx, dtdy;
  
  imin = ng;
  imax = ng+nx-1;

  jmin = ng;
  jmax = ng+ny-1;

  qx = 2*ng+nx;
  qy = 2*ng+ny;
    
  
  /* 
     start by computing the primitive variables -- we need to use these
     to compute the central differences normal to the interface. 
    
     To match the equations in Colella, we will assume the ordering
     V = (r, u, v, p, X_k)^T
  */

  r = dens;

  u_data = (double *) malloc(sizeof(double)*qx*qy);
  v_data = (double *) malloc(sizeof(double)*qx*qy);
  p_data = (double *) malloc(sizeof(double)*qx*qy);

  xn_data = (double *) malloc(sizeof(double)*qx*qy*nspecies);

  u = cArray_2d(u_data, qx, qy);
  v = cArray_2d(v_data, qx, qy);
  p = cArray_2d(p_data, qx, qy);

  xn = cArray_3d(xn_data, nspecies, qx, qy);

  zero_2d(u, qx, qy);
  zero_2d(v, qx, qy);
  zero_2d(p, qx, qy);
  zero_3d(xn, nspecies, qx, qy);

  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {

      ri = 1.0/r[i][j];

      u[i][j] = xmom[i][j]*ri;
      v[i][j] = ymom[i][j]*ri;
      
      p[i][j] = (ener[i][j] - 
		 0.5*dens[i][j]*(u[i][j]*u[i][j] + 
				 v[i][j]*v[i][j]))*(gamma - 1.0);

      for (n = 0; n < nspecies; n++) {
	xn[n][i][j] = species[n][i][j]*ri;
      }

    }
  }
  
  smallp = 1.e-10;


  /*--------------------------------------------------------------------------
    flattening
    --------------------------------------------------------------------------*/
    
  /* if we are doing flattening (Colella Eqs. 4.9 and 4.10), we should
     compute it here -- it needs to be direction independent.
  */

  xi_data = (double *) malloc(sizeof(double)*qx*qy);
  xi = cArray_2d(xi_data, qx, qy);
  zero_2d(xi, qx, qy);

  delta = 0.33;
  z0 = 0.75;
  z1 = 0.85;
    
  /* do each direction independently first */
  if (idir == 1) {
    _flatten(1, nx, ny, ng, p, u, smallp, delta, z0, z1, xi);
  } else if (idir == 2) {
    _flatten(2, nx, ny, ng, p, v, smallp, delta, z0, z1, xi);
  }


  /*--------------------------------------------------------------------------
    monotonized central differences
    --------------------------------------------------------------------------*/

  /* compute the limited slopes for the primitive variables */
  ldelta_r_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_u_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_v_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_p_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_xn_data = (double *) malloc(sizeof(double)*nspecies*qx*qy);

  ldelta_r = cArray_2d(ldelta_r_data, qx, qy);
  ldelta_u = cArray_2d(ldelta_u_data, qx, qy);
  ldelta_v = cArray_2d(ldelta_v_data, qx, qy);
  ldelta_p = cArray_2d(ldelta_p_data, qx, qy);
  ldelta_xn = cArray_3d(ldelta_xn_data, nspecies, qx, qy);
  
  zero_2d(ldelta_r, qx, qy);
  zero_2d(ldelta_u, qx, qy);
  zero_2d(ldelta_v, qx, qy);
  zero_2d(ldelta_p, qx, qy);
  zero_3d(ldelta_xn, nspecies, qx, qy);

  V_l_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  V_r_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  
  V_l = cArray_3d(V_l_data, NVAR+nspecies, qx, qy);
  V_r = cArray_3d(V_r_data, NVAR+nspecies, qx, qy);
  
  zero_3d(V_l, NVAR+nspecies, qx, qy);
  zero_3d(V_r, NVAR+nspecies, qx, qy);

  U_l_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  U_r_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));

  U_l = cArray_3d(U_l_data, NVAR+nspecies, qx, qy);
  U_r = cArray_3d(U_r_data, NVAR+nspecies, qx, qy);

  zero_3d(U_l, NVAR+nspecies, qx, qy);
  zero_3d(U_r, NVAR+nspecies, qx, qy);

  if (idir == 1) {

    /* do the x-direction first */
    _limit4(1, nx, ny, ng, r, ldelta_r);
    _limit4(1, nx, ny, ng, u, ldelta_u);
    _limit4(1, nx, ny, ng, v, ldelta_v);
    _limit4(1, nx, ny, ng, p, ldelta_p);

    for (n = 0; n < nspecies; n++) {
      _limit4(1, nx, ny, ng, xn[n], ldelta_xn[n]);
    }


    for (i = 0; i < qx; i++) {
      for (j = 0; j < qy; j++) {
	ldelta_r[i][j] = xi[i][j]*ldelta_r[i][j];
	ldelta_u[i][j] = xi[i][j]*ldelta_u[i][j];
	ldelta_v[i][j] = xi[i][j]*ldelta_v[i][j];
	ldelta_p[i][j] = xi[i][j]*ldelta_p[i][j];

	for (n = 0; n < nspecies; n++) {
	  ldelta_xn[n][i][j] = xi[i][j]*ldelta_xn[n][i][j];
	}

      }
    }
            

    /*
      We want to compute the left and right primitive variable states
      V = (r, u, v, p, X_k)^T
    */

    dtdx = dt/dx;

    _states(1, nx, ny, ng, nspecies,
	    gamma, dtdx,
	    r, u, v, p, xn,
	    ldelta_r, ldelta_u, ldelta_v, ldelta_p, ldelta_xn,
	    V_l[DENS], V_r[DENS],
	    V_l[XVEL], V_r[XVEL],
	    V_l[YVEL], V_r[YVEL],
	    V_l[PRES], V_r[PRES],
	    &(V_l[NVAR]), &(V_r[NVAR]));

    /* 
       now that we have the Vhats, transform them into conserved
       variables, U = (r, mx, my, re)
    */

    for (i = 0; i < qx; i++) {
      for (j = 0; j < qy; j++) {
	U_l[DENS][i][j] = V_l[DENS][i][j];
	U_l[XMOM][i][j] = V_l[DENS][i][j]*V_l[XVEL][i][j];
	U_l[YMOM][i][j] = V_l[DENS][i][j]*V_l[YVEL][i][j];
	U_l[ENER][i][j] = V_l[PRES][i][j]/(gamma - 1.0) + 
	  0.5*V_l[DENS][i][j]*(V_l[XVEL][i][j]*V_l[XVEL][i][j] +
			       V_l[YVEL][i][j]*V_l[YVEL][i][j]);

	for (n = 0; n < nspecies; n++) {
	  U_l[NVAR+n][i][j] = V_l[DENS][i][j]*V_l[NVAR+n][i][j];
	}


	U_r[DENS][i][j] = V_r[DENS][i][j];
	U_r[XMOM][i][j] = V_r[DENS][i][j]*V_r[XVEL][i][j];
	U_r[YMOM][i][j] = V_r[DENS][i][j]*V_r[YVEL][i][j];
	U_r[ENER][i][j] = V_r[PRES][i][j]/(gamma - 1.0) + 
	  0.5*V_r[DENS][i][j]*(V_r[XVEL][i][j]*V_r[XVEL][i][j] +
			       V_r[YVEL][i][j]*V_r[YVEL][i][j]);

	for (n = 0; n < nspecies; n++) {
	  U_r[NVAR+n][i][j] = V_r[DENS][i][j]*V_r[NVAR+n][i][j];
	}

      }
    }

  } else if (idir == 2) {

    /* now do the y-direction */

    _limit4(2, nx, ny, ng, r, ldelta_r);
    _limit4(2, nx, ny, ng, u, ldelta_u);
    _limit4(2, nx, ny, ng, v, ldelta_v);
    _limit4(2, nx, ny, ng, p, ldelta_p);

    for (n = 0; n < nspecies; n++) {
      _limit4(2, nx, ny, ng, xn[n], ldelta_xn[n]);
    }


    for (i = 0; i < qx; i++) {
      for (j = 0; j < qy; j++) {
	ldelta_r[i][j] = xi[i][j]*ldelta_r[i][j];
	ldelta_u[i][j] = xi[i][j]*ldelta_u[i][j];
	ldelta_v[i][j] = xi[i][j]*ldelta_v[i][j];
	ldelta_p[i][j] = xi[i][j]*ldelta_p[i][j];

	for (n = 0; n < nspecies; n++) {
	  ldelta_xn[n][i][j] = xi[i][j]*ldelta_xn[n][i][j];
	}

      }
    }
  
  

    /*
      We want to compute the left and right primitive variable states
      V = (r, u, v, p, X_k)^T
    */

    dtdy = dt/dy;

    _states(2, nx, ny, ng, nspecies,
	    gamma, dtdy,
	    r, u, v, p, xn,
	    ldelta_r, ldelta_u, ldelta_v, ldelta_p, ldelta_xn,
	    V_l[DENS], V_r[DENS],
	    V_l[XVEL], V_r[XVEL],
	    V_l[YVEL], V_r[YVEL],
	    V_l[PRES], V_r[PRES],
	    &(V_l[NVAR]), &(V_r[NVAR]));

    /*
      now that we have the Vhats, transform them into conserved
      variables, U = (r, mx, my, re)
    */
    for (i = 0; i < qx; i++) {
      for (j = 0; j < qy; j++) {
	U_l[DENS][i][j] = V_l[DENS][i][j];
	U_l[XMOM][i][j] = V_l[DENS][i][j]*V_l[XVEL][i][j];
	U_l[YMOM][i][j] = V_l[DENS][i][j]*V_l[YVEL][i][j];
	U_l[ENER][i][j] = V_l[PRES][i][j]/(gamma - 1.0) + 
	  0.5*V_l[DENS][i][j]*(V_l[XVEL][i][j]*V_l[XVEL][i][j] +
			       V_l[YVEL][i][j]*V_l[YVEL][i][j]);
	
	for (n = 0; n < nspecies; n++) {
	  U_l[NVAR+n][i][j] = V_l[DENS][i][j]*V_l[NVAR+n][i][j];
	}


	U_r[DENS][i][j] = V_r[DENS][i][j];
	U_r[XMOM][i][j] = V_r[DENS][i][j]*V_r[XVEL][i][j];
	U_r[YMOM][i][j] = V_r[DENS][i][j]*V_r[YVEL][i][j];
	U_r[ENER][i][j] = V_r[PRES][i][j]/(gamma - 1.0) + 
	  0.5*V_r[DENS][i][j]*(V_r[XVEL][i][j]*V_r[XVEL][i][j] +
			       V_r[YVEL][i][j]*V_r[YVEL][i][j]);

	for (n = 0; n < nspecies; n++) {
	  U_r[NVAR+n][i][j] = V_r[DENS][i][j]*V_r[NVAR+n][i][j];
	}

      }
    }

  }
  

  /*--------------------------------------------------------------------------
    apply the source terms
    --------------------------------------------------------------------------*/
  if (idir == 2) {

    for (i = 0; i < qx; i++) {
      for (j = 0; j < qy; j++) {

	U_l[YMOM][i][j] += 0.5*dt*dens[i][j-1]*grav;
	U_l[ENER][i][j] += 0.5*dt*ymom[i][j-1]*grav; 
	
	U_r[YMOM][i][j] += 0.5*dt*dens[i][j]*grav;
	U_r[ENER][i][j] += 0.5*dt*ymom[i][j]*grav; 

      }
    }

  }

  
  /*--------------------------------------------------------------------------
    solve the Riemann problem across the interface
    --------------------------------------------------------------------------*/
  zero_3d(F, NVAR+nspecies, qx, qy);

  if (idir == 1) {
    /*
    _HLLC(1, nx, ny, ng, nspecies,
	  gamma,
	  U_l[DENS], U_l[XMOM], U_l[YMOM], U_l[ENER], &(U_l[NVAR]),
	  U_r[DENS], U_r[XMOM], U_r[YMOM], U_r[ENER], &(U_r[NVAR]),
	  F[DENS], F[XMOM], F[YMOM], F[ENER], &(F[NVAR]));
    */
    _HLLC(1, nx, ny, ng,
	  gamma,
	  U_l[DENS], U_l[XMOM], U_l[YMOM], U_l[ENER],
	  U_r[DENS], U_r[XMOM], U_r[YMOM], U_r[ENER],
	  F[DENS], F[XMOM], F[YMOM], F[ENER]);

  } else if (idir == 2) {
    /*
    _HLLC(2, nx, ny, ng, nspecies,
	  gamma,
	  U_l[DENS], U_l[XMOM], U_l[YMOM], U_l[ENER], &(U_l[NVAR]),
	  U_r[DENS], U_r[XMOM], U_r[YMOM], U_r[ENER], &(U_r[NVAR]),
	  F[DENS], F[XMOM], F[YMOM], F[ENER], &(F[NVAR]));
    */
    _HLLC(2, nx, ny, ng, 
	  gamma,
	  U_l[DENS], U_l[XMOM], U_l[YMOM], U_l[ENER], 
	  U_r[DENS], U_r[XMOM], U_r[YMOM], U_r[ENER], 
	  F[DENS], F[XMOM], F[YMOM], F[ENER]);
  }


  /* clean up */
  free(u_data);  free_2d(u);
  free(v_data);  free_2d(v);
  free(p_data);  free_2d(p);
  free(xn_data);  free_3d(xn, nspecies);

  free(xi_data);  free_2d(xi);

  free(ldelta_r_data);  free_2d(ldelta_r);
  free(ldelta_u_data);  free_2d(ldelta_u);
  free(ldelta_v_data);  free_2d(ldelta_v);
  free(ldelta_p_data);  free_2d(ldelta_p);
  free(ldelta_xn_data);  free_3d(ldelta_xn, nspecies);

  free(V_l_data);  free_3d(V_l, NVAR+nspecies);
  free(V_r_data);  free_3d(V_r, NVAR+nspecies);

  free(U_l_data);  free_3d(U_l, NVAR+nspecies);
  free(U_r_data);  free_3d(U_r, NVAR+nspecies);

  return 1;
}

