#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pyro.h"
#include "interface.h"
#include "reconstruction.h"

/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------

Implementation of the Colella 2nd order unsplit Godunov scheme.  This
is a 2-dimensional implementation only.  We assume that the grid is
uniform, but it is relatively straightforward to relax this
assumption.

There are several different options for this solver (they are all
discussed in the Colella paper).

  limiter          = 1 to use the 2nd order MC limiter
                   = 2 to use the 4th order MC limiter

  hotf             = t to use the highorder reconstruction for the
                     transverse flux differences
                   = f to use a straight Godunov reconstruction for
                     the transverse flux difference

  use_flattening   = t to use the multidimensional flattening
                     algorithm at shocks

  delta, z0, z1      these are the flattening parameters.  The default
                     are the values listed in Colella 1990.

  
   j+3/2--+---------+---------+---------+
          |         |         |         |
     j+1 _|         |         |         |
          |         |         |         |
          |         |         |         |
   j+1/2--+---------XXXXXXXXXXX---------+
          |         X         X         |
       j _|         X         X         |
          |         X         X         |
          |         X         X         |
   j-1/2--+---------XXXXXXXXXXX---------+
          |         |         |         |
     j-1 _|         |         |         |
          |         |         |         |
          |         |         |         |
   j-3/2--+---------+---------+---------+
          |    |    |    |    |    |    |    
              i-1        i        i+1        
        i-3/2     i-1/2     i+1/2     i+3/2   

We wish to solve

  U_t + F^x_x + F^y_y = H


we want U_{i+1/2}^{n+1/2} -- the interface values that are input to
the Riemann problem through the faces for each zone.

Taylor expanding yields

   n+1/2                     dU           dU
  U          = U   + 0.5 dx  --  + 0.5 dt --
   i+1/2,j,L    i,j          dx           dt


                             dU             dF^x   dF^y
             = U   + 0.5 dx  --  - 0.5 dt ( ---- + ---- - H )
                i,j          dx              dx     dy


                              dU      dF^x            dF^y
             = U   + 0.5 ( dx -- - dt ---- ) - 0.5 dt ---- + 0.5 dt H
                i,j           dx       dx              dy


                                  dt       dU           dF^y
             = U   + 0.5 dx ( 1 - -- A^x ) --  - 0.5 dt ---- + 0.5 dt H
                i,j               dx       dx            dy


                                dt       _            dF^y
             = U   + 0.5  ( 1 - -- A^x ) DU  - 0.5 dt ---- + 0.5 dt H
                i,j             dx                     dy                


                     +----------+-----------+  +----+----+   +---+---+
                                |                   |            |
 
                    this is the monotonized   this is the   source term
                    central difference term   transverse 
                                              flux term
                                                  

There are two components, the central difference in the normal to
the interface, and the transverse flux difference.  This is done
for the left and right sides of all 4 interfaces in a zone, which
are then used as input to the Riemann problem, yielding the 1/2
time interface values,

     n+1/2
    U
     i+1/2,j

Then, the zone average values are updated in the usual finite-volume
way:

    n+1    n     dt    x  n+1/2       x  n+1/2     
   U    = U    + -- { F (U       ) - F (U       ) }
    i,j    i,j   dx       i-1/2,j        i+1/2,j

                 dt    y  n+1/2       y  n+1/2     
               + -- { F (U       ) - F (U       ) }
                 dy       i,j-1/2        i,j+1/2
    


Updating U_{i,j}:

  -- We want to find the state to the left and right (or top and
     bottom) of each interface, ex. U_{i+1/2,j,[lr]}^{n+1/2}, and use
     them to solve a Riemann problem across each of the four
     interfaces.

  -- U_{i+1/2,j,[lr]}^{n+1/2} is comprised of two parts, the
     computation of the monotonized central differences in the normal
     direction (eqs. 2.8, 2.10) and the computation of the transverse
     derivatives, which requires the solution of a Riemann problem in
     the transverse direction (eqs. 2.9, 2.14).

       -- the monotonized central difference part is computed using the
          primitive variables.

       -- We compute the central difference part in both directions
          before doing the transverse flux differencing, since for the
          high-order transverse flux implementation, we use these as
          the input to the transverse Riemann problem.
          
*/

/*===========================================================================
  unsplitFluxes
  =========================================================================== */

int _unsplitFluxes(int nx, int ny, int ng, int nspecies,
		   double dt, double dx, double dy, 
		   double cvisc, 
		   double gamma, double grav,
		   double **dens, double **xmom, double **ymom, double **ener,
		   double ***species,
		   double ***F_x, double ***F_y)
{
  
  /*
    unsplitFluxes returns the fluxes through the x and y interfaces by
    doing an unsplit reconstruction of the interface values and then
    solving the Riemann problem through all the interfaces at once
    
    currently we assume a gamma-law EOS
    
    grav is the gravitational acceleration in the y-direction
  */

  int i, j, n;

  int imin, imax;
  int jmin, jmax;
  int qx, qy;

  double **r;
  double ri;
  double **u, *u_data;
  double **v, *v_data;
  double **p, *p_data;
  double ***xn, *xn_data;

  double **xi, *xi_data;
  double **xi_x, *xi_x_data, **xi_y, *xi_y_data;

  double ***U_xl, *U_xl_data, ***U_xr, *U_xr_data;
  double ***U_yl, *U_yl_data, ***U_yr, *U_yr_data;

  double **ldelta_r, *ldelta_r_data;
  double **ldelta_u, *ldelta_u_data;
  double **ldelta_v, *ldelta_v_data;
  double **ldelta_p, *ldelta_p_data;
  double ***ldelta_xn, *ldelta_xn_data;

  double ***V_l, *V_l_data, ***V_r, *V_r_data;

  double delta, z0, z1, smallp;

  double dtdx, dtdy, dtdx2, dtdy2;

  double *avisco_x_data, **avisco_x, *avisco_y_data, **avisco_y;

  imin = ng;
  imax = ng+nx-1;

  jmin = ng;
  jmax = ng+ny-1;

  qx = 2*ng+nx;
  qy = 2*ng+ny;
    
  
  /* 
     start by computing the primitive variables -- we need to use these
     to compute the central differences normal to the interface. 
    
     To match the equations in Colella, we will assume the ordering
     V = (r, u, v, p, X_k)^T
  */


  r = dens;

  u_data = (double *) malloc(sizeof(double)*qx*qy);
  v_data = (double *) malloc(sizeof(double)*qx*qy);
  p_data = (double *) malloc(sizeof(double)*qx*qy);

  xn_data = (double *) malloc(sizeof(double)*qx*qy*nspecies);

  u = cArray_2d(u_data, qx, qy);
  v = cArray_2d(v_data, qx, qy);
  p = cArray_2d(p_data, qx, qy);
  
  xn = cArray_3d(xn_data, nspecies, qx, qy);

  zero_2d(u, qx, qy);
  zero_2d(v, qx, qy);
  zero_2d(p, qx, qy);
  zero_3d(xn, nspecies, qx, qy);

  smallp = 1.e-10;


  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {

      ri = 1.0/r[i][j];

      u[i][j] = xmom[i][j]*ri;
      v[i][j] = ymom[i][j]*ri;
      
      p[i][j] = (ener[i][j] - 
		 0.5*dens[i][j]*(u[i][j]*u[i][j] + 
				 v[i][j]*v[i][j]))*(gamma - 1.0);
      p[i][j] = max(smallp, p[i][j]);

      for (n = 0; n < nspecies; n++) {
	xn[n][i][j] = species[n][i][j]*ri;
      }

    }
  }
  

  /*--------------------------------------------------------------------------
    flattening
    --------------------------------------------------------------------------*/

    
  /* if we are doing flattening (Colella Eqs. 4.9 and 4.10), we should
     compute it here -- it needs to be direction independent.
  */

  xi_data = (double *) malloc(sizeof(double)*qx*qy);
  xi_x_data = (double *) malloc(sizeof(double)*qx*qy);
  xi_y_data = (double *) malloc(sizeof(double)*qx*qy);

  xi = cArray_2d(xi_data, qx, qy);
  xi_x = cArray_2d(xi_x_data, qx, qy);
  xi_y = cArray_2d(xi_y_data, qx, qy);

  zero_2d(xi, qx, qy);
  zero_2d(xi_x, qx, qy);
  zero_2d(xi_y, qx, qy);

  delta = 0.33;
  z0 = 0.75;
  z1 = 0.85;
    
  /* do each direction independently first */
  _flatten(1, nx, ny, ng, p, u, smallp, delta, z0, z1, xi_x);
  _flatten(2, nx, ny, ng, p, v, smallp, delta, z0, z1, xi_y);

  _flatten_multid(nx, ny, ng, p, xi_x, xi_y, xi);


  /*--------------------------------------------------------------------------
    monotonized central differences
    --------------------------------------------------------------------------*/

  /* compute the limited slopes for the primitive variables */
  ldelta_r_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_u_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_v_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_p_data = (double *) malloc(sizeof(double)*qx*qy);
  ldelta_xn_data = (double *) malloc(sizeof(double)*nspecies*qx*qy);

  ldelta_r = cArray_2d(ldelta_r_data, qx, qy);
  ldelta_u = cArray_2d(ldelta_u_data, qx, qy);
  ldelta_v = cArray_2d(ldelta_v_data, qx, qy);
  ldelta_p = cArray_2d(ldelta_p_data, qx, qy);
  ldelta_xn = cArray_3d(ldelta_xn_data, nspecies, qx, qy);
  
  zero_2d(ldelta_r, qx, qy);
  zero_2d(ldelta_u, qx, qy);
  zero_2d(ldelta_v, qx, qy);
  zero_2d(ldelta_p, qx, qy);
  zero_3d(ldelta_xn, nspecies, qx, qy);

  /* consider the x-direction first -- limit away */
  _limit4(1, nx, ny, ng, r, ldelta_r);
  _limit4(1, nx, ny, ng, u, ldelta_u);
  _limit4(1, nx, ny, ng, v, ldelta_v);
  _limit4(1, nx, ny, ng, p, ldelta_p);

  for (n = 0; n < nspecies; n++) {
    _limit4(1, nx, ny, ng, xn[n], ldelta_xn[n]);
  }
  

  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {
      ldelta_r[i][j] = xi[i][j]*ldelta_r[i][j];
      ldelta_u[i][j] = xi[i][j]*ldelta_u[i][j];
      ldelta_v[i][j] = xi[i][j]*ldelta_v[i][j];
      ldelta_p[i][j] = xi[i][j]*ldelta_p[i][j];

      for (n = 0; n < nspecies; n++) {
	ldelta_xn[n][i][j] = xi[i][j]*ldelta_xn[n][i][j];
      }

    }
  }
            

  /*
    We want to compute the left and right primitive variable states
    V = (r, u, v, p, X_k)^T
  */

  V_l_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  V_r_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));

  V_l = cArray_3d(V_l_data, NVAR+nspecies, qx, qy);
  V_r = cArray_3d(V_r_data, NVAR+nspecies, qx, qy);

  zero_3d(V_l, NVAR+nspecies, qx, qy);
  zero_3d(V_r, NVAR+nspecies, qx, qy);


  U_xl_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  U_xr_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));

  U_xl = cArray_3d(U_xl_data, NVAR+nspecies, qx, qy);
  U_xr = cArray_3d(U_xr_data, NVAR+nspecies, qx, qy);

  zero_3d(U_xl, NVAR+nspecies, qx, qy);
  zero_3d(U_xr, NVAR+nspecies, qx, qy);


  /* note -- we are assuming a uniform grid here */
  dtdx = dt/dx;
  dtdy = dt/dy;

  _states(1, nx, ny, ng, nspecies,
	  gamma, dtdx,
	  r, u, v, p, xn,
	  ldelta_r, ldelta_u, ldelta_v, ldelta_p, ldelta_xn,
	  V_l[DENS], V_r[DENS],
	  V_l[XVEL], V_r[XVEL],
	  V_l[YVEL], V_r[YVEL],
	  V_l[PRES], V_r[PRES],
	  &(V_l[NVAR]), &(V_r[NVAR]));

    
  /* 
     now that we have the Vhats, transform them into conserved
     variables, U = (r, mx, my, re)
  */

  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {
      U_xl[DENS][i][j] = V_l[DENS][i][j];
      U_xl[XMOM][i][j] = V_l[DENS][i][j]*V_l[XVEL][i][j];
      U_xl[YMOM][i][j] = V_l[DENS][i][j]*V_l[YVEL][i][j];
      U_xl[ENER][i][j] = V_l[PRES][i][j]/(gamma - 1.0) + 
	0.5*V_l[DENS][i][j]*(V_l[XVEL][i][j]*V_l[XVEL][i][j] +
			     V_l[YVEL][i][j]*V_l[YVEL][i][j]);
      
      for (n = 0; n < nspecies; n++) {
	U_xl[NVAR+n][i][j] = V_l[DENS][i][j]*V_l[NVAR+n][i][j];
      }


      U_xr[DENS][i][j] = V_r[DENS][i][j];
      U_xr[XMOM][i][j] = V_r[DENS][i][j]*V_r[XVEL][i][j];
      U_xr[YMOM][i][j] = V_r[DENS][i][j]*V_r[YVEL][i][j];
      U_xr[ENER][i][j] = V_r[PRES][i][j]/(gamma - 1.0) + 
	0.5*V_r[DENS][i][j]*(V_r[XVEL][i][j]*V_r[XVEL][i][j] +
			     V_r[YVEL][i][j]*V_r[YVEL][i][j]);

      for (n = 0; n < nspecies; n++) {
	U_xr[NVAR+n][i][j] = V_r[DENS][i][j]*V_r[NVAR+n][i][j];
      }

    }
  }

  /* do the y-direction */

  _limit4(2, nx, ny, ng, r, ldelta_r);
  _limit4(2, nx, ny, ng, u, ldelta_u);
  _limit4(2, nx, ny, ng, v, ldelta_v);
  _limit4(2, nx, ny, ng, p, ldelta_p);

  for (n = 0; n < nspecies; n++) {
    _limit4(2, nx, ny, ng, xn[n], ldelta_xn[n]);
  }


  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {
      ldelta_r[i][j] = xi[i][j]*ldelta_r[i][j];
      ldelta_u[i][j] = xi[i][j]*ldelta_u[i][j];
      ldelta_v[i][j] = xi[i][j]*ldelta_v[i][j];
      ldelta_p[i][j] = xi[i][j]*ldelta_p[i][j];

      for (n = 0; n < nspecies; n++) {
	ldelta_xn[n][i][j] = xi[i][j]*ldelta_xn[n][i][j];
      }

    }
  }
  
    
  /*
    We want to compute the left and right primitive variable states
    V = (r, u, v, p, X_k)^T
  */

  U_yl_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));
  U_yr_data = (double *) malloc(sizeof(double)*qx*qy*(NVAR+nspecies));

  U_yl = cArray_3d(U_yl_data, NVAR+nspecies, qx, qy);
  U_yr = cArray_3d(U_yr_data, NVAR+nspecies, qx, qy);

  zero_3d(V_l, NVAR+nspecies, qx, qy);
  zero_3d(V_r, NVAR+nspecies, qx, qy);

  zero_3d(U_yl, NVAR+nspecies, qx, qy);
  zero_3d(U_yr, NVAR+nspecies, qx, qy);

  _states(2, nx, ny, ng, nspecies, 
	  gamma, dtdy,
	  r, u, v, p, xn,
	  ldelta_r, ldelta_u, ldelta_v, ldelta_p, ldelta_xn,
	  V_l[DENS], V_r[DENS],
	  V_l[XVEL], V_r[XVEL],
	  V_l[YVEL], V_r[YVEL],
	  V_l[PRES], V_r[PRES],
	  &(V_l[NVAR]), &(V_r[NVAR]));


  /*
    now that we have the Vhats, transform them into conserved
    variables, U = (r, mx, my, re)
  */
  for (i = 0; i < qx; i++) {
    for (j = 0; j < qy; j++) {
      U_yl[DENS][i][j] = V_l[DENS][i][j];
      U_yl[XMOM][i][j] = V_l[DENS][i][j]*V_l[XVEL][i][j];
      U_yl[YMOM][i][j] = V_l[DENS][i][j]*V_l[YVEL][i][j];
      U_yl[ENER][i][j] = V_l[PRES][i][j]/(gamma - 1.0) + 
	0.5*V_l[DENS][i][j]*(V_l[XVEL][i][j]*V_l[XVEL][i][j] +
			  V_l[YVEL][i][j]*V_l[YVEL][i][j]);

      for (n = 0; n < nspecies; n++) {
	U_yl[NVAR+n][i][j] = V_l[DENS][i][j]*V_l[NVAR+n][i][j];
      }


      U_yr[DENS][i][j] = V_r[DENS][i][j];
      U_yr[XMOM][i][j] = V_r[DENS][i][j]*V_r[XVEL][i][j];
      U_yr[YMOM][i][j] = V_r[DENS][i][j]*V_r[YVEL][i][j];
      U_yr[ENER][i][j] = V_r[PRES][i][j]/(gamma - 1.0) + 
	0.5*V_r[DENS][i][j]*(V_r[XVEL][i][j]*V_r[XVEL][i][j] +
			     V_r[YVEL][i][j]*V_r[YVEL][i][j]);

      for (n = 0; n < nspecies; n++) {
	U_yr[NVAR+n][i][j] = V_r[DENS][i][j]*V_r[NVAR+n][i][j];
      }

    }
  }


  /*--------------------------------------------------------------------------
    apply the source terms
    --------------------------------------------------------------------------*/
  for (i = imin-2; i < imax+2; i++) {
    for (j = jmin-2; j < jmax+2; j++) {

      U_xl[YMOM][i][j] += 0.5*dt*dens[i-1][j]*grav;
      U_xl[ENER][i][j] += 0.5*dt*ymom[i-1][j]*grav;

      U_xr[YMOM][i][j] += 0.5*dt*dens[i][j]*grav;
      U_xr[ENER][i][j] += 0.5*dt*ymom[i][j]*grav;

      U_yl[YMOM][i][j] += 0.5*dt*dens[i][j-1]*grav;
      U_yl[ENER][i][j] += 0.5*dt*ymom[i][j-1]*grav; 

      U_yr[YMOM][i][j] += 0.5*dt*dens[i][j]*grav;
      U_yr[ENER][i][j] += 0.5*dt*ymom[i][j]*grav; 

    }
  }

  /*---------------------------------------------------------------------------
    do the transverse flux difference
    --------------------------------------------------------------------------*/

  /*
    solve the Riemann problem across each y-interface -- this is needed
    for the computation of the transverse flux difference
    
    we can do the low-order transverse flux (LOTF) approximation, which
    essentially just does a Godunov method for the input states to
    the Riemann problem, or we can do a high-order transverse flux
    (HOTF), which uses the reconstructions done above in the transverse
    direction as input to the Riemann problem.
  */


  zero_3d(F_x, NVAR+nspecies, qx, qy);
  zero_3d(F_y, NVAR+nspecies, qx, qy);

  _HLLC(1, nx, ny, ng, gamma,
	U_xl[DENS], U_xl[XMOM], U_xl[YMOM], U_xl[ENER], 
	U_xr[DENS], U_xr[XMOM], U_xr[YMOM], U_xr[ENER], 
	F_x[DENS], F_x[XMOM], F_x[YMOM], F_x[ENER]);

  _HLLC(2, nx, ny, ng, gamma,
	U_yl[DENS], U_yl[XMOM], U_yl[YMOM], U_yl[ENER], 
	U_yr[DENS], U_yr[XMOM], U_yr[YMOM], U_yr[ENER], 
	F_y[DENS], F_y[XMOM], F_y[YMOM], F_y[ENER]);

  
  /*--------------------------------------------------------------------------
    construct the interface values of U now
    --------------------------------------------------------------------------*/
  /*
    finally, we can construct the state perpendicular to the interface
    by adding the central difference part to the trasverse flux difference.
  */

  dtdx2 = 0.5*dtdx;
  dtdy2 = 0.5*dtdy;


  /*
    The states that we represent by indices i,j are shown below (1,2,3,4):
    
    
      j+3/2--+----------+----------+----------+
             |          |          |          |
             |          |          |          |             
        j+1 -+          |          |          |
             |          |          |          |
             |          |          |          |    1: U_xl[i,j,:] = U
      j+1/2--+----------XXXXXXXXXXXX----------+                      i-1/2,j,L
             |          X          X          |
             |          X          X          |             
          j -+        1 X 2        X          |    2: U_xr[i,j,:] = U
             |          X          X          |                      i-1/2,j,R
             |          X    4     X          |
      j-1/2--+----------XXXXXXXXXXXX----------+   
             |          |    3     |          |    3: U_yl[i,j,:] = U
             |          |          |          |                      i,j-1/2,L
        j-1 -+          |          |          |
             |          |          |          |
             |          |          |          |    4: U_yr[i,j,:] = U
      j-3/2--+----------+----------+----------+                      i,j-1/2,R
             |    |     |    |     |    |     |    
                 i-1         i         i+1        
           i-3/2      i-1/2      i+1/2      i+3/2   


    remember that the fluxes are stored on the left edge, so


    F_x[i,j,:] = F_x
                    i-1/2, j

    F_y[i,j,:] = F_y
                    i, j-1/2

  */
  
  for (n = 0; n < NVAR+nspecies; n++) {
    for (i = 1; i < qx-1; i++) {
      for (j = 1; j < qy-1; j++) {

	U_xl[n][i][j] = U_xl[n][i][j] - dtdy2*(F_y[n][i-1][j+1] - F_y[n][i-1][j]);
	U_xr[n][i][j] = U_xr[n][i][j] - dtdy2*(F_y[n][i][j+1] - F_y[n][i][j]);
	U_yl[n][i][j] = U_yl[n][i][j] - dtdx2*(F_x[n][i+1][j-1] - F_x[n][i][j-1]);
	U_yr[n][i][j] = U_yr[n][i][j] - dtdx2*(F_x[n][i+1][j] - F_x[n][i][j]);

      }
    }
  }

  /*
    up until now, F_x and F_y stored the transverse fluxes, now we zero it
    out and use it to store the normal fluxes
  */

  zero_3d(F_x, NVAR+nspecies, qx, qy);
  zero_3d(F_y, NVAR+nspecies, qx, qy);

  _HLLC(1, nx, ny, ng, gamma,
	U_xl[DENS], U_xl[XMOM], U_xl[YMOM], U_xl[ENER], 
	U_xr[DENS], U_xr[XMOM], U_xr[YMOM], U_xr[ENER], 
	F_x[DENS], F_x[XMOM], F_x[YMOM], F_x[ENER]);

  _HLLC(2, nx, ny, ng, gamma,
	U_yl[DENS], U_yl[XMOM], U_yl[YMOM], U_yl[ENER], 
	U_yr[DENS], U_yr[XMOM], U_yr[YMOM], U_yr[ENER], 
	F_y[DENS], F_y[XMOM], F_y[YMOM], F_y[ENER]);


  /* -----------------------------------------------------------------------
     compute the artifical viscosity
     ----------------------------------------------------------------------- */
  avisco_x_data = (double *) malloc(sizeof(double)*qx*qy);
  avisco_y_data = (double *) malloc(sizeof(double)*qx*qy);

  avisco_x = cArray_2d(avisco_x_data, qx, qy);
  avisco_y = cArray_2d(avisco_y_data, qx, qy);

  artifical_viscosity(nx, ny, ng,
                      cvisc,
                      dx, dy,
                      u, v,
                      avisco_x, avisco_y);


  /* -----------------------------------------------------------------------
     update the fluxes with the artifical viscosity
     ----------------------------------------------------------------------- */
  for (i = imin-1; i <= imax+1; i++) {
    for (j = jmin-1; j <= jmax+1; j++) {

      F_x[DENS][i][j] = F_x[DENS][i][j] + avisco_x[i][j]*(dens[i-1][j] - dens[i][j]);
      F_x[XMOM][i][j] = F_x[XMOM][i][j] + avisco_x[i][j]*(xmom[i-1][j] - xmom[i][j]);
      F_x[YMOM][i][j] = F_x[YMOM][i][j] + avisco_x[i][j]*(ymom[i-1][j] - ymom[i][j]);
      F_x[ENER][i][j] = F_x[ENER][i][j] + avisco_x[i][j]*(ener[i-1][j] - ener[i][j]);


      F_y[DENS][i][j] = F_y[DENS][i][j] + avisco_y[i][j]*(dens[i][j-1] - dens[i][j]);
      F_y[XMOM][i][j] = F_y[XMOM][i][j] + avisco_y[i][j]*(xmom[i][j-1] - xmom[i][j]);
      F_y[YMOM][i][j] = F_y[YMOM][i][j] + avisco_y[i][j]*(ymom[i][j-1] - ymom[i][j]);
      F_y[ENER][i][j] = F_y[ENER][i][j] + avisco_y[i][j]*(ener[i][j-1] - ener[i][j]);

    }
  }


  /* clean up */
  free(u_data);  free_2d(u);
  free(v_data);  free_2d(v);
  free(p_data);  free_2d(p);
  free(xn_data);  free_3d(xn, nspecies);

  free(xi_data);  free_2d(xi);
  free(xi_x_data);  free_2d(xi_x);
  free(xi_y_data);  free_2d(xi_y);

  free(ldelta_r_data);  free_2d(ldelta_r);
  free(ldelta_u_data);  free_2d(ldelta_u);
  free(ldelta_v_data);  free_2d(ldelta_v);
  free(ldelta_p_data);  free_2d(ldelta_p);
  free(ldelta_xn_data);  free_3d(ldelta_xn, nspecies);

  free(V_l_data);  free_3d(V_l, NVAR+nspecies);
  free(V_r_data);  free_3d(V_r, NVAR+nspecies);
  
  free(U_xl_data);  free_3d(U_xl, NVAR+nspecies);
  free(U_xr_data);  free_3d(U_xr, NVAR+nspecies);
  free(U_yl_data);  free_3d(U_yl, NVAR+nspecies);
  free(U_yr_data);  free_3d(U_yr, NVAR+nspecies);

  free(avisco_x_data);  free_2d(avisco_x);
  free(avisco_y_data);  free_2d(avisco_y);

  return 1;
}

