/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Python.h"
#include "numarray/numarray.h"
#include "numarray/libnumarray.h"
#include "pyro.h"
#include "hdf5.h"

int store_double(char *name, hid_t group, double *data);
int store_integer(char *name, hid_t group, int *data);

int read_double(char *name, hid_t group, double *data);
int read_integer(char *name, hid_t group, int *data);


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

/* open up the hdf5 file for output and return the file handle. */
static PyObject *open_file(PyObject *self, PyObject *args) {

  hid_t file_id;
  char *filename;

  if (!PyArg_ParseTuple(args, "s", &filename))
    return NULL;
 
  file_id = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);

  return Py_BuildValue("i", (int) file_id);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

/* open up the hdf5 file for reading and return the file handle. */
static PyObject *open_for_read(PyObject *self, PyObject *args) {

  hid_t file_id;

  char *filename;

  if (!PyArg_ParseTuple(args, "s", &filename))
    return NULL;
 
  file_id = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);

  return Py_BuildValue("i", (int) file_id);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

/* close the HDF5 file given the file identifier */
static PyObject *close_file(PyObject *self, PyObject *args) {

  int file_id;
  int ierr;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;
 
  ierr = H5Fclose((hid_t) file_id);

  return Py_BuildValue("i", ierr);
}
  

/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

/* Write the header to the HDF5 file, storing the data as attributes.
   The expected arguments are the file identifier (as returned from the 
   open_file routine), the list of variables, the simulation time, 
   the timestep, and the step number

   write_header(file_id, vars, time, dt, nsteps, nx, ny, 
                xmin, xmax, ymin, ymax, bcs)
*/

static PyObject *write_header(PyObject *self, PyObject *args) {

  PyObject *vars, *bcs;
  double time, dt;
  int nsteps;
  int nx, ny;

  double xmin, xmax, ymin, ymax;

  int nvars, nbcs;
 
  int status;
  
  PyObject *item;

  int n;

  char about_message[] = "pyro @VERSION@ output";

  char svars[MAX_VARS][MAX_VARNAME_LENGTH];
  char sbcs[NBCS][MAX_BCNAME_LENGTH];

  int file_id;

  hid_t group_id;  
  hid_t attribute, attribute_space;

  hid_t string_type;

  int rank;
  hsize_t dimens, dimens2d[2];

  if (!PyArg_ParseTuple(args, "iO!ddiiiddddO!", 
			&file_id,
			&PyList_Type, &vars,
			&time, &dt,
			&nsteps, 
			&nx, &ny,
			&xmin, &xmax, &ymin, &ymax,
			&PyList_Type, &bcs) )
    return NULL;

  status = 1;


  /* ----------------------------------------------------------------------
      get the number of variable names and copy them into the svars array
     ---------------------------------------------------------------------- */

  nvars = PyList_Size(vars);

  if (nvars > MAX_VARS) {
    printf("ERROR: too many variables defined\n");
    printf("       increase MAX_VARS and rebuild the io_util wrappers\n");
    return Py_BuildValue("i", -1);
  }
    
  for (n = 0; n < nvars; n++) {
    item = PyList_GetItem(vars, n);

    if (!PyString_Check(item)) {
      printf("ERROR: element %d of the variable list is not a string\n", n);
      return Py_BuildValue("i", -1);
    }

    strncpy(svars[n], PyString_AsString(item), MAX_VARNAME_LENGTH);

    if (PyString_Size(item) > MAX_VARNAME_LENGTH) {
      printf("WARNING: variable %d name (%s) too long\n", n, svars[n]);
      printf("         name being truncated.  Increase MAX_VARNAME_LENGTH\n");
    }
 
  }


  /* ----------------------------------------------------------------------
      get the BC names
     ---------------------------------------------------------------------- */

  nbcs = PyList_Size(bcs);

  if (nbcs != NBCS) {
    printf("ERROR: incorrect number of BCs defined\n");
    return Py_BuildValue("i", -1);
  }
    
  for (n = 0; n < nbcs; n++) {
    item = PyList_GetItem(bcs, n);

    if (!PyString_Check(item)) {
      printf("ERROR: element %d of the BC list is not a string\n", n);
      return Py_BuildValue("i", -1);
    }

    strncpy(sbcs[n], PyString_AsString(item), MAX_BCNAME_LENGTH);

    if (PyString_Size(item) > MAX_BCNAME_LENGTH) {
      printf("WARNING: BC %d name (%s) too long\n", n, sbcs[n]);
      printf("         name being truncated.  Increase MAX_BCNAME_LENGTH\n");
    }
 
  }


  /* ----------------------------------------------------------------------
      store the data
     ---------------------------------------------------------------------- */
  group_id = H5Gopen((hid_t) file_id, "/");
  if (group_id < 0) {
    printf("ERROR openning group / in file\n");
    status = -1;
  }

  rank = 1;
  dimens = 1;

  /* ----- about this file ------------------------------------------------ */
  string_type = H5Tcopy(H5T_C_S1);
  if (string_type < 0) {
    printf("ERROR defining string_type in file\n");
    status = -1;
  }

  H5Tset_size(string_type, strlen(about_message)+1);

  attribute_space = H5Screate_simple(rank, &dimens, NULL);
  attribute = H5Acreate(group_id, "about this file", string_type,
			attribute_space, H5P_DEFAULT);

  if (H5Awrite(attribute, string_type, about_message) < 0) {
    printf("ERROR writing variables about message to file\n");
    status = -1;
  }
  
  H5Tclose(string_type);
  H5Sclose(attribute_space);
  H5Aclose(attribute);

  /* time, timestep, and number of steps */
  status = store_double("time", group_id, &time);
  status = store_double("dt", group_id, &dt);

  status = store_integer("number of steps", group_id, &nsteps);


  /* domain size */
  status = store_integer("nx", group_id, &nx);
  status = store_integer("ny", group_id, &ny);


  /* ----- domain extrema -------------------------------------------------- */
  status = store_double("xmin", group_id, &xmin);
  status = store_double("xmax", group_id, &xmax);
  status = store_double("ymin", group_id, &ymin);
  status = store_double("ymax", group_id, &ymax);


  /* ----- variables ------------------------------------------------------ */
  rank = 2;

  dimens2d[0] = nvars;
  dimens2d[1] = 1;

  string_type = H5Tcopy(H5T_C_S1);
  H5Tset_size(string_type, MAX_VARNAME_LENGTH);

  attribute_space = H5Screate_simple(rank, dimens2d, NULL);
  attribute = H5Acreate(group_id, "variables", string_type,
			attribute_space, H5P_DEFAULT);

  if (H5Awrite(attribute, string_type, svars) < 0) {
    printf("ERROR writing variables attribute to file\n");
    status = -1;
  }
  
  H5Tclose(string_type);
  H5Sclose(attribute_space);
  H5Aclose(attribute);


  /* ----- BC names -------------------------------------------------------- */
  rank = 2;

  dimens2d[0] = nbcs;
  dimens2d[1] = 1;

  string_type = H5Tcopy(H5T_C_S1);
  H5Tset_size(string_type, MAX_BCNAME_LENGTH);

  attribute_space = H5Screate_simple(rank, dimens2d, NULL);
  attribute = H5Acreate(group_id, "boundary conditions", string_type,
			attribute_space, H5P_DEFAULT);

  if (H5Awrite(attribute, string_type, sbcs) < 0) {
    printf("ERROR writing boundary conditions attribute to file\n");
    status = -1;
  }
  
  H5Tclose(string_type);
  H5Sclose(attribute_space);
  H5Aclose(attribute);
 

  H5Gclose(group_id);
  
  return Py_BuildValue("i", status);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

/* Write the data to the HDF5 file, a single variable at a time.
   The expected arguments are the file identifier (as returned from the 
   open_file routine), the variable name, and the data */
static PyObject *write_data(PyObject *self, PyObject *args) {

  PyObject *odata;
  PyArrayObject *data;

  char *varname;

  int file_id;

  int status;

  hid_t dataspace, dataset, memspace;

  int rank;
  hsize_t dimens2d[2];


  if (!PyArg_ParseTuple(args, "isO", &file_id, &varname, &odata))
    return NULL;

  data = (PyArrayObject *) NA_InputArray(odata, tFloat64, NUM_C_ARRAY);

  if (!data || data->nd != 2) {
    PyErr_SetString(PyExc_ValueError,
		    "write_data: error converting input array");
     return NULL;  
  }

  rank = 2;
  dimens2d[0] = data->dimensions[0];
  dimens2d[1] = data->dimensions[1];

  dataspace = H5Screate_simple(rank, dimens2d, NULL);

  dataset = H5Dcreate((hid_t) file_id, varname, H5T_NATIVE_DOUBLE,
                      dataspace, H5P_DEFAULT); 

  memspace = H5Screate_simple(rank, dimens2d, NULL);
  
  status = H5Dwrite(dataset, H5T_NATIVE_DOUBLE, memspace, dataspace,
                    H5P_DEFAULT, data->data);

  Py_XDECREF(data);

  H5Sclose(memspace);
  H5Sclose(dataspace);
  H5Dclose(dataset);

  return Py_BuildValue("i", status);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_simulation_time(PyObject *self, PyObject *args) {

  /* time, dt = io_util.readSimulationTime(fileId) */

  double time, dt;
 
  int file_id;
  hid_t group_id;  

  int status;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");

  status = read_double("time", group_id, &time);
  status = read_double("dt", group_id, &dt);

  H5Gclose(group_id);

  return Py_BuildValue("dd", time, dt);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_simulation_step(PyObject *self, PyObject *args) {

  /* nstep = io_util.readSimulationTime(fileId) */

  int nstep;
 
  int file_id;
  hid_t group_id;  


  int status;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");

  status = read_integer("number of steps", group_id, &nstep);

  H5Gclose(group_id);

  return Py_BuildValue("i", nstep);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_grid_size(PyObject *self, PyObject *args) {

  /* nx, ny = io_util.readSimulationTime(fileId) */

  int nx, ny;
 
  int file_id;
  hid_t group_id;  

  int status;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");

  status = read_integer("nx", group_id, &nx);
  status = read_integer("ny", group_id, &ny);

  H5Gclose(group_id);

  return Py_BuildValue("ii", nx, ny);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_domain_extrema(PyObject *self, PyObject *args) {

  /* xmin, xmax, ymin, ymax = io_util.readDomainExtrema(fileId) */

  double xmin, xmax, ymin, ymax;
 
  int file_id;
  hid_t group_id;  

  int status;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");

  status = read_double("xmin", group_id, &xmin);
  status = read_double("xmax", group_id, &xmax);
  status = read_double("ymin", group_id, &ymin);
  status = read_double("ymax", group_id, &ymax);

  H5Gclose(group_id);

  return Py_BuildValue("dddd", xmin, xmax, ymin, ymax);
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_var_list(PyObject *self, PyObject *args) {

  /* vars = io_util.readSimulationTime(fileId) */
  
  char svars[MAX_VARS][MAX_VARNAME_LENGTH];

  int nvar;

  int file_id;

  hid_t group_id;  
  hid_t attribute, attribute_space;
  hid_t string_type;

  hsize_t dims[2], maxdims[2];

  int i;

  PyObject *var_list;
  PyObject *string;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");


  attribute = H5Aopen_name(group_id, "variables");
  attribute_space = H5Aget_space(attribute);

  string_type = H5Aget_type(attribute);

  /* get the number of variables by looking at the size of the dataspace */
  H5Sget_simple_extent_dims(attribute_space, dims, maxdims);
  nvar = dims[0];


  if (H5Aread(attribute, string_type, svars) < 0) {
    printf("ERROR reading variables attribute from file\n");
    return NULL;
  }
  
  H5Aclose(attribute);
  H5Sclose(attribute_space);
  H5Tclose(string_type);

  /* make a new python list object containing the variables */
  var_list = PyList_New(nvar);
  
  for (i = 0; i < nvar; i++) {
    string = PyString_FromString(svars[i]);
    PyList_SetItem(var_list, i, string);
  }

  H5Gclose(group_id);

  return var_list;
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *read_bc_list(PyObject *self, PyObject *args) {

  /* vars = io_util.readSimulationTime(fileId) */
  
  char sbcs[NBCS][MAX_BCNAME_LENGTH];

  int nbcs;

  int file_id;

  hid_t group_id;  
  hid_t attribute, attribute_space;
  hid_t string_type;

  hsize_t dims[2], maxdims[2];

  int i;

  PyObject *bc_list;
  PyObject *string;

  if (!PyArg_ParseTuple(args, "i", &file_id))
    return NULL;

  group_id = H5Gopen((hid_t) file_id, "/");

  attribute = H5Aopen_name(group_id, "boundary conditions");
  attribute_space = H5Aget_space(attribute);

  string_type = H5Aget_type(attribute);

  /* get the number of variables by looking at the size of the dataspace */
  H5Sget_simple_extent_dims(attribute_space, dims, maxdims);
  nbcs = dims[0];


  if (H5Aread(attribute, string_type, sbcs) < 0) {
    printf("ERROR reading BCs attribute from file\n");
    return NULL;
  }
  
  H5Aclose(attribute);
  H5Sclose(attribute_space);
  H5Tclose(string_type);

  /* make a new python list object containing the variables */
  bc_list = PyList_New(nbcs);
  
  for (i = 0; i < nbcs; i++) {
    string = PyString_FromString(sbcs[i]);
    PyList_SetItem(bc_list, i, string);
  }

  H5Gclose(group_id);

  return bc_list;
}


static PyObject *read_data(PyObject *self, PyObject *args) {

  /* we are going to read in the dimensions of the data and then 
     create a new numarray array and store the data into that. */

  static PyObject *odata;
  double *data;

  char *varname;

  int nx, ny;
  int file_id;

  hid_t dataspace, dataset;

  hsize_t dims[2], maxdims[2];


  if (!PyArg_ParseTuple(args, "is", &file_id, &varname))
    return NULL;

  dataset = H5Dopen((hid_t) file_id, varname);
  dataspace = H5Dget_space(dataset);

  H5Sget_simple_extent_dims(dataspace, dims, maxdims);
  nx = dims[0];
  ny = dims[1];


  data = (double *) malloc(nx*ny*sizeof(double));
  if (data == NULL) {
    printf("ERROR: unable to allocate data array in read_data\n");
    return NULL;
  }
  
  if (H5Dread(dataset, H5T_NATIVE_DOUBLE, H5S_ALL, 
	      dataspace, H5P_DEFAULT, data) < 0) {
    printf("ERROR reading in dataset %s\n", varname);
    return NULL;
  }

  odata = (PyObject *) NA_NewArray((void *) data, tFloat64, 2, nx, ny);

  free (data);

  H5Sclose(dataspace);
  H5Dclose(dataset);

  return odata;
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */  

static PyMethodDef writemethods[] = {
  {"openFile", open_file, METH_VARARGS},
  {"openForRead", open_for_read, METH_VARARGS},
  {"closeFile", close_file, METH_VARARGS},
  {"writeHeader", write_header, METH_VARARGS},
  {"writeData", write_data, METH_VARARGS},
  {"readSimulationTime", read_simulation_time, METH_VARARGS},
  {"readSimulationStep", read_simulation_step, METH_VARARGS},
  {"readGridSize", read_grid_size, METH_VARARGS},
  {"readDomainExtrema", read_domain_extrema, METH_VARARGS},
  {"readVarList", read_var_list, METH_VARARGS},
  {"readBCList", read_bc_list, METH_VARARGS},
  {"readData", read_data, METH_VARARGS},
  {NULL, NULL}
};


void initio_util(void) {
  Py_InitModule("io_util", writemethods);
  import_libnumarray();
}


int store_double(char *name, hid_t group, double *data) {

  int rank;
  hsize_t dimens;
  hid_t attribute, attribute_space;

  rank = 1;
  dimens = 1;

  attribute_space = H5Screate_simple(rank, &dimens, NULL);
  attribute = H5Acreate(group, name, H5T_NATIVE_DOUBLE,
			attribute_space, H5P_DEFAULT);
  
  if (H5Awrite(attribute, H5T_NATIVE_DOUBLE, data) < 0) {
    printf("ERROR writing %s attribute to file\n", name);
    return -1;
  }
  
  H5Sclose(attribute_space);
  H5Aclose(attribute);
  
  return 1;
}


int store_integer(char *name, hid_t group, int *data) {

  int rank;
  hsize_t dimens;
  hid_t attribute, attribute_space;

  rank = 1;
  dimens = 1;

  attribute_space = H5Screate_simple(rank, &dimens, NULL);
  attribute = H5Acreate(group, name, H5T_NATIVE_INT,
			attribute_space, H5P_DEFAULT);
  
  if (H5Awrite(attribute, H5T_NATIVE_INT, data) < 0) {
    printf("ERROR writing %s attribute to file\n", name);
    return -1;
  }
  
  H5Sclose(attribute_space);
  H5Aclose(attribute);
  
  return 1;
}


int read_double(char *name, hid_t group, double *data) {

  hid_t attribute;

  attribute = H5Aopen_name(group, name);
  if (H5Aread(attribute, H5T_NATIVE_DOUBLE, data) < 0) {
    printf("ERROR reading %s attribute from file\n", name);
    return -1;
  }

  H5Aclose(attribute);

  return 1;
}


int read_integer(char *name, hid_t group, int *data) {

  hid_t attribute;

  attribute = H5Aopen_name(group, name);
  if (H5Aread(attribute, H5T_NATIVE_INT, data) < 0) {
    printf("ERROR reading %s attribute from file\n", name);
    return -1;
  }

  H5Aclose(attribute);

  return 1;
}
