from numarray import *

import reconstructionlib
from util import runparams

"""
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
"""


def _restrict_cc(nx_f, ny_f, ng_f, U_f, nx_c, ny_c, ng_c, U_c):

    imin_f = ng_f
    imax_f = ng_f + nx_f - 1

    jmin_f = ng_f
    jmax_f = ng_f + ny_f - 1    

    imin_c = ng_c
    imax_c = ng_c + nx_c - 1

    jmin_c = ng_c
    jmax_c = ng_c + ny_c - 1    

    # fill the coarse patch with the restricted data -- just average
    # the fine zones into their corresponding coarse zone.  This is
    # done by shifting our view into the U_f array and using a stride
    # of 2 in the indexing.
    U_c[imin_c:imax_c+1,jmin_c:jmax_c+1] = \
                  0.25*(U_f[imin_f:imax_f+1:2,jmin_f:jmax_f+1:2] +
                        U_f[imin_f+1:imax_f+1:2,jmin_f:jmax_f+1:2] +
                        U_f[imin_f:imax_f+1:2,jmin_f+1:jmax_f+1:2] +
                        U_f[imin_f+1:imax_f+1:2,jmin_f+1:jmax_f+1:2])



def _prolong_cc(nx_c, ny_c, ng_c, U_c, nx_f, ny_f, ng_f, U_f):
    """    
    We will reconstruct the data in the zone from the zone-averaged
    variables using the same limited slopes as in the advection
    routine.  Getting a good multidimensional reconstruction
    polynomial is hard -- we want it to be bilinear and monotonic
    -- we settle for having each slope be independently monotonic:

              (x)         (y)
    f(x,y) = m    x/dx + m    y/dy + <f>

    where the m's are the limited differences in each direction.  When
    averaged over the parent cell, this reproduces <f>.

    Each zone's reconstrution will be averaged over 4 children.


    +-----------+     +-----+-----+
    |           |     |     |     |
    |           |     |  3  |  4  |
    |    <f>    | --> +-----+-----+              
    |           |     |     |     |
    |           |     |  1  |  2  |
    +-----------+     +-----+-----+                           
    
    We will fill each of the finer resolution zones by filling
    all the 1's together, using a stride 2 into the fine array.
    Then the 2's and ..., this allows us to operate in a vector
    fashion.  All operations will use the same slopes for their
    respective parents.
    
    """

    imin_c = ng_c
    imax_c = ng_c + nx_c - 1

    jmin_c = ng_c
    jmax_c = ng_c + ny_c - 1    

    imin_f = ng_f
    imax_f = ng_f + nx_f - 1

    jmin_f = ng_f
    jmax_f = ng_f + ny_f - 1

    # get the limited differences in each direction for all interior
    # zones in our current mesh
    m_x = zeros((nx_c + 2*ng_c, ny_c + 2*ng_c), Float64)
    m_y = zeros((nx_c + 2*ng_c, ny_c + 2*ng_c), Float64)
    
    limiter = 1

    m_x = reconstructionlib.limita(1, limiter, nx_c, ny_c, ng_c, U_c)
    m_y = reconstructionlib.limita(2, limiter, nx_c, ny_c, ng_c, U_c)            

    # fill the '1' children
    U_f[imin_f:imax_f+1:2,jmin_f:jmax_f+1:2] = \
                    U_c[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    - 0.25*m_x[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    - 0.25*m_y[imin_c:imax_c+1,jmin_c:jmax_c+1]


    # fill the '2' children
    U_f[imin_f+1:imax_f+1:2,jmin_f:jmax_f+1:2] = \
                    U_c[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    + 0.25*m_x[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    - 0.25*m_y[imin_c:imax_c+1,jmin_c:jmax_c+1]


    # fill the '3' children
    U_f[imin_f:imax_f+1:2,jmin_f+1:jmax_f+1:2] = \
                    U_c[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    - 0.25*m_x[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    + 0.25*m_y[imin_c:imax_c+1,jmin_c:jmax_c+1]


    # fill the '4' children
    U_f[imin_f+1:imax_f+1:2,jmin_f+1:jmax_f+1:2] = \
                    U_c[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    + 0.25*m_x[imin_c:imax_c+1,jmin_c:jmax_c+1] \
                    + 0.25*m_y[imin_c:imax_c+1,jmin_c:jmax_c+1]
            
    

class cellCentered:
    """
    the 2-d patch class.  The patch object will contain the coordinate
    information (at various centerings) and the unknown data itself.

    a patch is built in a multi-step process before it can be used:

    create the patch object:

        myPatch = patch.cellCentered(nx, ny)

    register any variables that w eexpect to live on this patch

        myPatch.registerVar('density')
        myPatch.registerVar('x-momentum')
        myPatch.registerVar('y-momentum')
        ...

    finally, initialize the patch

        myPatch.init()

    This last step actually allocates the storage for the unknowns.  Once
    this is done, the patch is considered to be locked.  New variables cannot
    be added.
    """

    # the constructor function.  The only data that we require is the number
    # of points that make up the mesh.
    #
    # We optionally take the extrema of the domain (assume it is [0,1]x[0,1])
    # and the number of guardcells (assume 1)

    def __init__ (self, nx, ny, ng=1, \
                  xmin=0.0, xmax=1.0, ymin=0.0, ymax=1.0, \
                  xlb = "reflect", xrb = "reflect", \
                  ylb = "reflect", yrb = "reflect"):

        self.allVars = []
        self.nAllVars = 0
        
        self.vars = []
        self.nvar = 0

        self.species = []
        self.nspecies = 0
        
        self.nx = nx
        self.ny = ny
        self.ng = ng

        # domain extrema
        self.xmin = xmin
        self.xmax = xmax

        self.ymin = ymin
        self.ymax = ymax

        # domain boundary conditions
        self.xlb = xlb
        self.xrb = xrb

        self.ylb = ylb
        self.yrb = yrb

        # compute the indices of the block interior (excluding guardcells)
        self.imin = ng
        self.imax = ng+nx-1

        self.jmin = ng
        self.jmax = ng+ny-1

        # define the coordinate information at the left, center, and right
        # zone positions
        self.dx = (xmax - xmin)/nx

        self.xl = (arange(nx+2*ng) - ng)*self.dx + xmin
        self.xr = (arange(nx+2*ng) + 1.0 - ng)*self.dx + xmin
        self.x = 0.5*(self.xl + self.xr)

        self.dy = (ymax - ymin)/ny

        self.yl = (arange(ny+2*ng) - ng)*self.dy + ymin
        self.yr = (arange(ny+2*ng) + 1.0 - ng)*self.dy + ymin
        self.y = 0.5*(self.yl + self.yr)

        # time/timestep information
        self.time = 0.0
        self.dt = 0.0
        self.nstep = 0
        

    def registerVar(self, name):
        """ register a variable with patch object """

        self.vars.append(name)
        self.nvar += 1


    def registerSpecies(self, name):
        """ register a variable with patch object """

        self.species.append(name)
        self.nspecies += 1


    def init(self):
        """
        called after all the variables are registered and allocates
        the storage for the unknowns
        """

        self.allVars = self.vars + self.species
        self.nAllVars = self.nvar + self.nspecies
        
        self.data = zeros((self.nAllVars, 2*self.ng+self.nx, 2*self.ng+self.ny),
                          Float64)

        
    def __str__(self):
        """ print out some basic information about the patch object """
        string = "2-d patch: nx = " + `self.nx` + ', ny = ' + `self.ny` + ', ng = ' + `self.ng` + "\n" + \
                 "           nAllVars = " + `self.nAllVars` + "\n" + \
                 "           all variables: " + `self.allVars`
        return string
        

    def getVar(self, name):

        n = self.allVars.index(name)
        # we have a choice here, we can return in essence, a pointer to the
        # variable in the main self.data array, or, by commenting out the
        # next line, return a copy, which would require an explicit put
        # method.
        #return self.data[:,:,n].copy()
        return self.data[n,:,:]


    def getVarByIndex(self, index):
        return self.data[index,:,:]


    def zero(self, name):
        n = self.allVars.index(name)
        self.data[n,:,:] = 0.0
        
        
    def prolong(self, varList=["all"]):
        """ create a new mesh object from the existing one by prolonging to
            a grid that is twice as large.
            """


        # create the new mesh object
        fP = cellCentered(self.nx*2, self.ny*2, ng=self.ng,
                          xmin=self.xmin, xmax=self.xmax,
                          ymin=self.ymin, ymax=self.ymax,
                          xlb=self.xlb, xrb=self.xrb,
                          ylb=self.ylb, yrb=self.yrb)


        try: varList.index("all")
        except ValueError:
            rvars = varList
        else:
            rvars = self.allVars
            
        for var in rvars:    
            fP.registerVar(var)

        fP.init()
        

        
        for var in rvars:    

            nself = self.allVars.index(var)
            nfp = rvars.index(var)

            _prolong_cc(self.nx, self.ny, self.ng, self.data[nself,:,:],
                        fP.nx, fP.ny, fP.ng, fP.data[nfp,:,:])
            
        return fP
    

        
    def restrict(self, varList=["all"]):
        """ restrict the current mesh to a coarser mesh.  Assume a factor of
            two in their resolution """

        # we can only restrict if our mesh is an even size
        if (self.nx%2 != 0 or self.ny%2 != 0):
            print "ERROR: restriction only operates on meshes with even sizes"
            return NULL
        
        # create the new mesh object
        cP = cellCentered(self.nx/2, self.ny/2, ng=self.ng,
                          xmin=self.xmin, xmax=self.xmax,
                          ymin=self.ymin, ymax=self.ymax,
                          xlb=self.xlb, xrb=self.xrb,
                          ylb=self.ylb, yrb=self.yrb)


        try: varList.index("all")
        except ValueError:
            rvars = varList
        else:
            rvars = self.allVars

        for var in rvars:    
            cP.registerVar(var)


        cP.init()

        for var in rvars:    

            nself = self.allVars.index(var)
            ncp = rvars.index(var)
            
            _restrict_cc(self.nx, self.ny, self.ng, self.data[nself,:,:],
                         cP.nx, cP.ny, cP.ng, cP.data[ncp,:,:])
            
        return cP
    

        
    def fillBC(self):
        """ fill the boundary conditions """
    
        # do the boundaries one at a time.  The single grid case
        # is easy -- every boundary is a physical boundary of the
        # domain, so there is not much trickery to do

        # we do periodic, outflow, reflect, and in the y-direction,
        # HSE boundaries.
        
        # lower X boundary
        if (self.xlb == "reflect" or self.xlb == "outflow"):
            n = 0
            while n < self.nvar:

                i = 0
                while i < self.imin:
                    self.data[n,i,:] = self.data[n,self.imin,:]
                    
                    i += 1
                
                n+= 1

            # if we are reflecting, actually reflect the x-momentum now
            if (self.xlb == "reflect"):
                n = self.allVars.index("x-momentum")

                i = 0
                while i < self.imin:
                    self.data[n,i,:] = -self.data[n,2*self.ng-i-1,:]
                    
                    i += 1


        elif (self.xlb == "periodic"):
            n = 0
            while n < self.nvar:

                i = 0
                while i < self.imin:
                    self.data[n,i,:] = self.data[n,self.imax-self.ng+i+1,:]

                    i += 1

                n += 1
            

        # upper X boundary
        if (self.xrb == "reflect" or self.xrb == "outflow"):
            n = 0
            while n < self.nvar:

                i = self.imax+1
                while i < self.nx+2*self.ng:
                    self.data[n,i,:] = self.data[n,self.imax,:]

                    i += 1
                
                n+= 1

            # if we are reflecting, actually reflect the x-momentum now
            if (self.xrb == "reflect"):
                n = self.allVars.index("x-momentum")

                i = self.imax+1
                while i < self.nx+2*self.ng:
                    self.data[n,i,:] = -self.data[n,2*self.imax-i+1+self.ng,:]
                    
                    i += 1


        elif (self.xrb == "periodic"):
            n = 0
            while n < self.nvar:

                i = self.imax+1
                while i < 2*self.ng + self.nx:
                    self.data[n,i,:] = self.data[n,i-self.imax-1+self.ng,:]

                    i += 1

                n += 1
                

        # lower Y boundary
        if (self.ylb == "reflect" or self.ylb == "outflow"):
            n = 0
            while n < self.nvar:

                j = 0
                while j < self.jmin:
                    self.data[n,:,j] = self.data[n,:,self.jmin]
                    
                    j += 1
                
                n+= 1

            # if we are reflecting, actually reflect the y-momentum now
            if (self.ylb == "reflect"):
                n = self.allVars.index("y-momentum")

                j = 0
                while j < self.jmin:
                    self.data[n,:,j] = -self.data[n,:,2*self.ng-j-1]
                    
                    j += 1


        elif (self.ylb == "periodic"):
            n = 0
            while n < self.nvar:

                j = 0
                while j < self.jmin:
                    self.data[n,:,j] = self.data[n,:,self.jmax-self.ng+j+1]

                    j += 1

                n += 1
                

        # upper Y boundary
        if (self.yrb == "reflect" or self.yrb == "outflow"):
            n = 0
            while n < self.nvar:

                j = self.jmax+1
                while j < self.ny+2*self.ng:
                    self.data[n,:,j] = self.data[n,:,self.jmax]
                    
                    j += 1
                
                n+= 1

            # if we are reflecting, actually reflect the y-momentum now
            if (self.yrb == "reflect"):
                n = self.allVars.index("y-momentum")

                j = self.jmax+1
                while j < self.ny+2*self.ng:
                    self.data[n,:,j] = -self.data[n,:,2*self.jmax-j+1]
                    
                    j += 1

        
        elif (self.yrb == "periodic"):
            n = 0
            while n < self.nvar:

                j = self.jmax+1
                while j < 2*self.ng + self.ny:
                    self.data[n,:,j] = self.data[n,:,j-self.jmax-1+self.ng]

                    j += 1

                n += 1

