"""
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
"""

"""

initialize the sedov problem.  This is a point explosion problem that
has an analytic solution (Sedov 1959).  Here we place E = 1 at a
single point, with the density constant but the surrounding pressure
very low (1.e-5).

In practice, we initialize the point by setting the pressure
corresponding to the energy deposition inside a radius r_init via

P = 3 (gamma - 1) E / ((vu + 1) pi r_init^nu)

where nu = 2 for a cylinder (a 2-d point explosion) and nu = 3 for a
spherical explosion.

"""

import sys
from util import runparams
from mesh import *
from numarray import *

def fillPatch(myPatch):
    """ initialize the sedov problem """

    print "initializing the sedov problem..."

    # make sure that we are passed a valid patch object
    if not isinstance(myPatch, patch.cellCentered):
        print "ERROR: patch invalid in io.dumpPatch()"
        print myPatch.__class__
        sys.exit()

    # get the density, momenta, and energy as separate variables
    dens = myPatch.getVar("density")
    xmom = myPatch.getVar("x-momentum")
    ymom = myPatch.getVar("y-momentum")
    ener = myPatch.getVar("energy")

    # initialize the components, remember, that ener here is
    # rho*eint + 0.5*rho*v**2, where eint is the specific
    # internal energy (erg/g)
    dens[:,:] = 1.0
    xmom[:,:] = 0.0
    ymom[:,:] = 0.0

    E_sedov = 1.0
    r_init = runparams.getParam("sedov.r_init")

    gamma = runparams.getParam("eos.gamma")
    pi = 3.14159
    
    xmin = runparams.getParam("mesh.xmin")
    xmax = runparams.getParam("mesh.xmax")

    ymin = runparams.getParam("mesh.ymin")
    ymax = runparams.getParam("mesh.ymax")

    xctr = 0.5*(xmin + xmax)
    yctr = 0.5*(ymin + ymax)
    
    # there is probably an easier way to do this, but for now, we
    # will just do an explicit loop.  Also, we really want to set
    # the pressue and get the internal energy from that, and then
    # compute the total energy (which is what we store).  For now
    # we will just fake this.
    nsub = 4
    
    i = myPatch.imin
    while i <= myPatch.imax:

        j = myPatch.jmin
        while j <= myPatch.jmax:

            dist = sqrt((myPatch.x[i] - xctr)**2 +
                        (myPatch.y[j] - yctr)**2)
            if (dist < 2.0*r_init):
                pzone = 0.0
            
                ii = 0
                while ii < nsub:

                    jj = 0
                    while jj < nsub:

                        xsub = myPatch.xl[i] + (myPatch.dx/nsub)*(ii + 0.5)
                        ysub = myPatch.yl[j] + (myPatch.dy/nsub)*(jj + 0.5) 
                    
                        dist = sqrt((xsub - xctr)**2 + \
                                    (ysub - yctr)**2)

                        if dist <= r_init:
                            p = (gamma - 1.0)*E_sedov/(pi*r_init*r_init)
                        else:
                            p = 1.e-5

                        pzone += p

                        jj += 1
                    ii += 1

                p = pzone/(nsub*nsub)
            else:
                p = 1.e-5
                
            ener[i][j] = p/(gamma - 1.0)
            
            j += 1
        i += 1
        
    

    
                             
