#!/usr/bin/env python

"""
88d888b. dP    dP 88d888b. .d8888b.
88'  `88 88    88 88'  `88 88'  `88
88.  .88 88.  .88 88       88.  .88
88Y888P' `8888P88 dP       `88888P'
88            .88
dP        d8888P                     


  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale

  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
"""

usage = """
    pyro is a simple python framework for solving computational hydro
    problems.  At the moment, it is 2-d only.  This is mostly because
    it is not really meant as a production code, but rather as a simple
    code to gain familarity with different hydro methods.  2-d is usually
    enough for these purposes.

    usage:

      ./pyro [options] <solver name> <problem name> <input file>


    available options:

      -i
        automatically generate an inputs file, named inputs.auto with
        comments describing the available parameters

      -r file
        restart the calculation from file

      -t file
        compare the present solution to the previously stored test file


    arguments:
          
      <solver name>
        the name of the solver to use for the calculation.  At present
        this is either compressible or advection

      <problem name>
        the name of the problem we are solving, as defined in the
        problems/ directory for each different solver.  This defines
        the initial conditions and loads any problem specific runtime
        parameters

      <input file>
        the name of the inputs file, specifying the runtime parameters
        and their defaults.  An automatically generated version of this
        can be produced with the -i version
    """


"""
  see VERSION for current version number
  
  history 

  0.51  bug fix in gravity states and 4th order limiter
        some code clean-ups

  0.50  add gravity to the compressible solver.
        Add the bubble problem.
        Fix the reflecting boundary conditions.
        Clean up the compressible interface states routine to make it easier
            to add species.
        Fix the IDL read routine to deal with row-major/column-major issues
        Get the 4th order limiter working.
 
  0.42  add prolongation routine and mesh_test.py test routine
        initialize C return arrays to 0

  0.41  add automated testing script
        move reconstruction routines to a new module

  0.40  more centralized build system.
        finished the splitting of C routines into interface +
           implementation, now that Riemann is done.

  0.30  split the states C routines into interface and implementation
           routines for easier maintainence.
        flatten is now in C.
        bring the Strang split compressible solver back in sync with
           the unsplit version.

  0.25  added restart capability
        automagic generation of inputs file
        added testing

  0.20  reduced memory overhead
        greatly sped up the unsplit/compressible solver by making better
           use of python slices instead of loops, and modifying the Riemann
           solver interface.

  0.18  moved the states computation into C

  0.15  conversion from Numeric to numarray

  0.12  break out each solver into their own directory
        (i.e. compressible/, advection/)

  0.1   initial version
  
"""

# load the pyro modules
from io import *
from mesh import *
from util import runparams, physconst, profile, compare

# load the python modules
import sys, getopt
from numarray import *

# get started...
f = open("VERSION")
version = f.read()
version = version[:-1]
f.close()
print 'pyro ' + version + '...'

# parse the runtime arguments.  We specify the solver (which we import locally
# under the namespace 'solver', the problem name, and the input file name

if len(sys.argv) == 1:
    print usage
    sys.exit(2)



    
autoparamfile = 0
restartFile = ""
testFile = ""

try:
    opts, next = getopt.getopt(sys.argv[1:], "ir:t:")
    
except getopt.GetoptError:
    print "invalid calling sequence"
    sys.exit(2)
            

for o, a in opts:

    if o == "-i":
        autoparamfile = 1

    if o == "-r":
        restartFile = a

    if o == "-t":
        testFile = a
                                                  
try: solverName = next[0]
except IndexError:
    print 'ERROR: solver name not specified on command line'
    print 'USAGE: pyro.py [options] <solver name> <problem name> <input file>'
    sys.exit(2)

exec 'import ' + solverName + ' as solver'
exec 'from problems.' + solverName + ' import *'

try: problemName = next[1]
except IndexError:
    print 'ERROR: problem name not specified on command line'
    print 'USAGE: pyro.py [options] <solver name> <problem name> <input file>'
    sys.exit(2)

try: paramFile = next[2]
except IndexError:
    print 'ERROR: parameter file not specified on command line'
    print 'USAGE: pyro.py [options] <solver name> <problem name> <input file>'
    sys.exit(2)    



#-----------------------------------------------------------------------------
# set the parameter defaults
#-----------------------------------------------------------------------------

runparams.LoadParams("_defaults")

# set the problem specific defaults
print 'loading: ' + "problems/"+solverName+"/_"+problemName+".defaults"
runparams.LoadParams("problems/"+solverName+"/_"+problemName+".defaults")

# read in the parameter file and set the runtime parameters -- anything in the
# paramFile will overwrite the parameters initialized through the
# _problem.defaults file and the master defaults list
runparams.LoadParams(paramFile)


print 'runtime parameters: '
runparams.PrintAllParams()

if autoparamfile == 1:
    runparams.PrintParamFile()


init_tstep_factor = runparams.getParam("driver.init_tstep_factor")
max_dt_change = runparams.getParam("driver.max_dt_change")



# if we are starting from scratch, then we need to create a grid object
# and then initialize it with the problem specific initial conditions
if restartFile == "":
    
    #-------------------------------------------------------------------------
    # initialize the grid
    #-------------------------------------------------------------------------

    xmin = runparams.getParam("mesh.xmin")
    xmax = runparams.getParam("mesh.xmax")

    ymin = runparams.getParam("mesh.ymin")
    ymax = runparams.getParam("mesh.ymax")

    xlb = runparams.getParam("mesh.xl_boundary")
    xrb = runparams.getParam("mesh.xr_boundary")

    ylb = runparams.getParam("mesh.yl_boundary")
    yrb = runparams.getParam("mesh.yr_boundary")

    nx = runparams.getParam("mesh.nx")
    ny = runparams.getParam("mesh.ny")

    myPatch = patch.cellCentered(nx, ny, ng=4, \
                                 xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                                 xlb=xlb, xrb=xrb, ylb=ylb, yrb=yrb)

    # each solver dictates the variables that it must know about -- load
    # them now.
    solver.initialize(myPatch)

    # initialize the problem specific variables and species
    #exec problemName + '.initialize(myPatch)'
    
    myPatch.init()

    print myPatch


    #-------------------------------------------------------------------------
    # initialize the problem
    #-------------------------------------------------------------------------

    # initialize the grid with our desired problem.  The problem name was a
    # runtime input, but each problem is guaranteed to have an initialize()
    # method, which serves as the only entry point into the problem.
    print 'looking to initialize using problem: ', problemName
    exec problemName + '.fillPatch(myPatch)'

    myPatch.time = 0.0
    myPatch.nstep = 0

    myPatch.fillBC()

    # get the initial timestep
    solver.timestep(myPatch)
    myPatch.dt = init_tstep_factor*myPatch.dt
    dtOld = myPatch.dt
    


else:
    myPatch = io.readFile(restartFile)
    dtOld = myPatch.dt
    
    myPatch.fillBC()




#-----------------------------------------------------------------------------
# evolve
#-----------------------------------------------------------------------------

max_steps = runparams.getParam("driver.max_steps")
max_time = runparams.getParam("driver.max_time")


# dump out the initial file
filename = runparams.getParam("io.basename") + "%4.4d" % myPatch.nstep
io.writeFile(filename, myPatch)

nplot = runparams.getParam("io.nplot")

while (myPatch.nstep < max_steps and myPatch.time < max_time):

    timerLoop = profile.timer('timestep loop')
    timerLoop.begin()
    
    # evolve through a single timestep
    err = solver.evolve(myPatch)
    if (err == -1):
        print "ERROR: fatal error in evolution"
        sys.exit(2)

        
    # set the boundary conditions
    myPatch.fillBC()

    
    # get the new timestep
    solver.timestep(myPatch)
    myPatch.dt = minimum(myPatch.dt, max_dt_change*dtOld)

    if  myPatch.time + myPatch.dt > max_time:
        myPatch.dt = max_time - myPatch.time
        
    dtOld = myPatch.dt

    
    myPatch.time += myPatch.dt
    myPatch.nstep += 1

    print "%5d:  %12.7e   %12.7e" % (myPatch.nstep, myPatch.time, myPatch.dt)

    if (myPatch.nstep % nplot == 0):
        filename = runparams.getParam("io.basename") + "%4.4d" % myPatch.nstep
        io.writeFile(filename, myPatch)


    timerLoop.end()
    

    
# do a final output -- only if we didn't already do it above
if (myPatch.nstep % nplot != 0):
    filename = runparams.getParam("io.basename") + "%4.4d" % myPatch.nstep
    io.writeFile(filename, myPatch)

profile.timeReport()


# compare the solution to the stored test file if we are testing
if testFile != "":

    errors = 0
    
    testPatch = io.readFile(testFile)
    
    compare.mesh2d(myPatch,testPatch)

    
