/*
  ---------------------------------------------------------------------
  Copyright (C) 2003, 2004  Michael Zingale
   
  This software is distributed under the terms of the GNU General
  Public License, version 2.  See COPYING for details.  It is covered
  by no warantee whatsoever, explicit or implied.
  ---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
 
#include "Python.h"
#include "numarray/numarray.h"
#include "numarray/libnumarray.h"

#include "pyro.h"
#include "pyroarray.h"
#include "reconstruction.h"   

/* wrappers to pass information back and forth between python and C */

static PyObject * limita(PyObject *self, PyObject *args) {

  PyObject *oa, *ola;
  PyArrayObject *pa;

  int idir, itype;
  int nx, ny, ng;
  int qx, qy;

  double **a;
  double *ldelta_a_data;
  double **ldelta_a;

  int ierr; 

  /* parse the inputs:
     
     limita(idir, nx, ny, ng, a)

     idir  = direction to limit in; 1 = x, 2 = y
     nx/ny = number of zones in the x/y-direction
     ng    = number of guardcells
     a     = array of the variable we are limiting
  */

  if (!PyArg_ParseTuple(args, "iiiiiO",
			&idir, &itype, &nx, &ny, &ng,
			&oa))
    return NULL;

  qx = 2*ng + nx;
  qy = 2*ng + ny;

  py2C_init();
  a = py2C_create2d(oa, qx, qy, &pa);
    

  /* allocate space for the return */
  ldelta_a_data = (double *) malloc(sizeof(double)*qx*qy);
  if ((ldelta_a = cArray_2d(ldelta_a_data, qx, qy)) == NULL) {
    printf("ERROR in limita: unable to allocate memory\n");
    return NULL;
  }
  zero_2d(ldelta_a, qx, qy);


  /* actually do the limiting now */
  if (itype == 1) {
    ierr = _limit(idir, nx, ny, ng, a, ldelta_a);
  } else if (itype == 2) {
    ierr = _limit4(idir, nx, ny, ng, a, ldelta_a);
  } else {
    printf("ERROR: invalid limiter type\n");
    return NULL;
  }

  if (ierr < 0) {
    printf("ERROR in limita: illegal return value from _limit\n");
    return NULL;
  }

  /* create the return PyObject */
  ola = (PyObject *) NA_NewArray((void *) ldelta_a_data, tFloat64, 2, qx, qy);
   

  /* free up memory */
  free(ldelta_a_data);
  free_2d(ldelta_a);
  py2C_free2d(a, qx, qy, pa);

  return ola;
}


/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *flattena(PyObject *self, PyObject *args) {
 
  /* a vectorized version of the flattening algorithm */
   
  PyObject *op, *ou, *oxi;
  PyArrayObject *pp, *pu;
 
  double **p, **u;
 
  double smallp, delta, z0, z1;

  /* we will create an array to pass the data back */
  double *xi_data;
  double **xi;
 
  int idir;
  int nx, ny, ng;
  int qx, qy;

  int ierr;

  /* parse the inputs:

     flattena(idir, nx, ny, ng, p, u, smallp, delta, z0, z1)

     idir   = direction to compute the flattening coefficent in
     nx/ny  = number of zones in the x/y-direction
     ng     = number of guardcells
     p      = pressure field
     u      = velocity field (x-velocity if idir=1, y-velocity if idir=2)
     smallp = small pressure cutoff
     delta  = pressure jump threshold
     z0, z1 = flattening parameters
  */
 
  if (!PyArg_ParseTuple(args, "iiiiOOdddd",
                        &idir, &nx, &ny, &ng,
                        &op, &ou,
                        &smallp, &delta, &z0, &z1))
    return NULL;

  qx = 2*ng + nx;
  qy = 2*ng + ny;

  py2C_init();
  p = py2C_create2d(op, qx, qy, &pp);
  u = py2C_create2d(ou, qx, qy, &pu);

 
  /* allocate space for the return */
  xi_data = (double *) malloc(sizeof(double)*qx*qy);
  if ((xi = cArray_2d(xi_data, qx, qy)) == NULL) {
    printf("ERROR in flattena: unable to allocate memory\n");
    return NULL;
  }
  zero_2d(xi, qx, qy);
      

  /* actually do the flattening now */
  ierr = _flatten(idir, nx, ny, ng, p, u, smallp, delta, z0, z1, xi);
  if (ierr < 0) {
    printf("ERROR in flattena: illegal return value from _flatten\n");
    return NULL;
  }


  free (xi);
  py2C_free2d(p, qx, qy, pp);
  py2C_free2d(u, qx, qy, pu);
   
  oxi = (PyObject *) NA_NewArray((void *) xi_data, tFloat64, 2, qx, qy);
  free(xi_data);
 
  return oxi; 
}
 

/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyObject *flatten_multid(PyObject *self, PyObject *args) {
 
  /* Take the flattening coefficient in each direction and merge to 
     produce a single coefficient */
   
  PyObject *op, *oxi_x, *oxi_y, *oxi;
  PyArrayObject *pp, *pxi_x, *pxi_y;
 
  double **p, **xi_x, **xi_y;
 
  /* we will create an array to pass the data back */
  double *xi_data;
  double **xi;
 
  int nx, ny, ng;
  int qx, qy;

  int ierr;

  /* parse the inputs:

     flatten_multid(nx, ny, ng, p, xi_x, xi_y)

     nx/ny     = number of zones in the x-direction
     ny        = number of zones in the y-direction
     ng        = number of guardcells
     p         = pressure field
     xi_x/xi_y = flattening coefficient in x/y-direction
  */
 
  if (!PyArg_ParseTuple(args, "iiiOOO",
                        &nx, &ny, &ng,
                        &op, &oxi_x, &oxi_y))
    return NULL;

  qx = 2*ng + nx;
  qy = 2*ng + ny;

  py2C_init();
  p = py2C_create2d(op, qx, qy, &pp);
  xi_x = py2C_create2d(oxi_x, qx, qy, &pxi_x);
  xi_y = py2C_create2d(oxi_y, qx, qy, &pxi_y);

 
  /* allocate space for the return */
  xi_data = (double *) malloc(sizeof(double)*qx*qy);
  if ((xi = cArray_2d(xi_data, qx, qy)) == NULL) {
    printf("ERROR in flattena: unable to allocate memory\n");
    return NULL;
  }
  zero_2d(xi, qx, qy);
      

  /* actually do the flattening now */
  ierr = _flatten_multid(nx, ny, ng, p, xi_x, xi_y, xi);
  if (ierr < 0) {
    printf("ERROR in flatten_multid: illegal return value from _flatten\n");
    return NULL;
  }


  free_2d(xi);
  py2C_free2d(p, qx, qy, pp);
  py2C_free2d(xi_x, qx, qy, pxi_x);
  py2C_free2d(xi_y, qx, qy, pxi_y);
   
  oxi = (PyObject *) NA_NewArray((void *) xi_data, tFloat64, 2, qx, qy);
  free(xi_data);
 
  return oxi;
}
 

/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

static PyMethodDef reconstructionmethods[] = {
  {"limita", limita, METH_VARARGS},
  {"flattena", flattena, METH_VARARGS},
  {"flatten_multid", flatten_multid, METH_VARARGS},
  {NULL, NULL}
};

 
void initreconstructionlib(void) {
  Py_InitModule("reconstructionlib", reconstructionmethods);
  import_libnumarray();
}
 
